"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pycity_base.classes.demand.apartment as apm

from pycity_scheduling.classes.thermal_entity_cooling import ThermalEntityCooling
from pycity_scheduling.classes.thermal_entity_heating import ThermalEntityHeating
from pycity_scheduling.classes.electrical_entity import ElectricalEntity
from pycity_scheduling.classes.entity_container import EntityContainer


class Apartment(EntityContainer, apm.Apartment):
    """
    Extension of pyCity_base class Apartment for scheduling purposes.

    Parameters
    ----------
    environment : Environment
        Common to all other objects. Includes time and weather instances
    net_floor_area : float, optional
        netto floor area in [m^2]
    occupancy : Occupancy, optional
        Occupancy profile from pycity_base for the apartment.

    Notes
    -----
    - Apartments inherit their set of constraints from EntityContainer.
    """

    def __init__(self, environment, net_floor_area=None, occupancy=None):
        super().__init__(environment, net_floor_area, occupancy)
        self._long_id = "APM_" + self._id_string

        self.th_cooling_demand_list = []
        self.th_heating_demand_list = []
        self.el_demand_list = []
        self.all_demand_list = []

    def addEntity(self, entity):
        """
        Add entity to apartment.

        Parameters
        ----------
        entity : OptimizationEntity
            Entity to be added to the apartment; must be of type FixedLoad,
            DeferrableLoad, CurtailableLoad, SpaceHeating, SpaceCooling or DomesticHotWater.
        """
        super(Apartment, self).addEntity(entity)

        if isinstance(entity, ThermalEntityCooling):
            self.th_cooling_demand_list.append(entity)
        if isinstance(entity, ThermalEntityHeating):
            self.th_heating_demand_list.append(entity)
        if isinstance(entity, ElectricalEntity):
            self.el_demand_list.append(entity)
        self.all_demand_list.append(entity)
        return

    def get_lower_entities(self):
        yield from self.all_demand_list



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import random
import os

from pycity_scheduling.classes import *
from pycity_scheduling.data.tabula_data import tabula_building_data as tbd
from pycity_scheduling.data.ev_data import ev_data as evd


def generate_standard_environment(**timer_args):
    """
    Generate a standard environment object.

    Returns
    -------
    pycity_scheduling.classes.Environment
    """
    timer = Timer(**timer_args)

    weather_file_name = "TRY_2018.dat"
    root_path = os.path.dirname(os.path.dirname(__file__))
    weather_file_path = os.path.join(root_path, "data", weather_file_name)
    weather = Weather(timer, path_TRY=weather_file_path, location=(51.53, 7.21), altitude=10.0)

    prices = Prices(timer)

    environment = Environment(timer, weather, prices)
    return environment


def _calculate_ev_times(timer):
    """
    Supporting function to calculate electric vehicle charging time slots.
    """
    length = int(3600/timer.time_discretization)
    ev_time_ranges = [
        [0] * (8 * length) + [1] * (12 * length) + [0] * (4 * length),
        [1] * (12 * length) + [0] * (12 * length),
        [0] * (12 * length) + [1] * (12 * length),
        [1] * (10 * length) + [0] * (12 * length) + [1] * (2 * length),
        [1] * (9 * length) + [0] * (12 * length) + [1] * (3 * length),
        [1] * (8 * length) + [0] * (12 * length) + [1] * (4 * length),
        [1] * (7 * length) + [0] * (12 * length) + [1] * (5 * length),
        [1] * (6 * length) + [0] * (12 * length) + [1] * (6 * length),
        [1] * (5 * length) + [0] * (12 * length) + [1] * (7 * length),
    ]
    return ev_time_ranges


def _calculate_dl_times(timer):
    """
    Supporting function to calculate deferrable load consumption time slots.
    """
    length = int(3600/timer.time_discretization)
    dl_time_ranges = [
        [1] * (8 * length) + [0] * (16 * length),
        [0] * (8 * length) + [1] * (4 * length) + [0] * (12 * length),
        [0] * (12 * length) + [1] * (4 * length) + [0] * (8 * length),
        [0] * (16 * length) + [1] * (4 * length) + [0] * (4 * length),
        [0] * (20 * length) + [1] * (4 * length),
        [0] * (17 * length) + [1] * (4 * length) + [0] * (3 * length),
        [0] * (7 * length) + [1] * (4 * length) + [0] * (13 * length),
        [0] * (10 * length) + [1] * (4 * length) + [0] * (10 * length),
        [0] * (2 * length) + [1] * (4 * length) + [0] * (18 * length),
    ]
    return dl_time_ranges


def _distribute(d, number):
    """
    Generates amounts based on distribution and total number and minimizes rounding error

    Parameters
    ----------
    d : dict
        The distribution used for generating amounts.
        Keys : str
        Values : float
            Number between 0 and 1. The sum over all values must be one.

    Returns
    ----------
    dict :
        dict with keys from original dict and amount as values
    """
    if any(not (0 <= v <= 1) for v in d.values()):
        raise ValueError("Bad device probabilities")
    amount_dict = {}
    # calculate the difference between sum of amounts and number when rounding down
    diff = number - sum(int(v * number) for v in d.values())
    if diff < 0 or diff > len(d):
        raise ValueError("Bad device probabilities")
    # sort dict by int round error
    sorted_distribution = sorted(d.items(), key=lambda item: (number * item[1]) % 1, reverse=True)
    # calculate amount for keys in distribution (keys with largest rounding error first)
    for i, (building, share) in enumerate(sorted_distribution):
        amount = int(share * number)
        if i < diff:
            # add 1 of key for keys with large rounding error to account for diff
            amount += 1
        amount_dict[building] = amount
    return amount_dict


def generate_tabula_buildings(environment,
                              number,
                              building_distribution=None,
                              heating_distribution=None,
                              device_probabilities=None,
                              objective='price',
                              seed=None,
                              mpi_interface=None):
    """
    Generate buildings based on the TABULA data.

    Generate buildings based on the TABULA data from: http://www.episcope.eu/ (accessed on 2020/09/28)
    Heating units are automatically dimensioned and added to each building. A
    THS always covers the thermal energy demand of a building for at least two
    hours.

    Parameters
    ----------
    environment : pycity_scheduling.classes.Environment
    number : int
        Number of houses to be generated.
    building_distribution : dict, optional
        The distribution of the houses among the tabula standard buildings. If
        omitted an equal distribution will be used.
        Keys : str
            'DE.N.<SFH|MFH>.<n>.Gen' or '<SFH|MFH>.<year>'
        Values : float
            Number between 0 and 1. The sum over all values must be one.
    heating_distribution : dict, optional
        The distribution of heating devices among the houses. If omitted an
        equal distribution will be used.
        Keys : str
            {'HP', 'EH', 'CHP', 'BL'}
        Values : float
            Number between 0 and 1. The sum over all values must be one.
    device_probabilities : dict, optional
        The probabilities of the houses / apartments to have the given device.
        Keys : str
            {'FL', 'DL', 'EV', 'PV', 'BAT'}
        Values : float
            Number between 0 and 1.
    objective : str, optional
        The objective for all buildings.
    seed: int, optional
        Specify a seed for the randomization. If omitted, a non-deterministic
        city district will be generated.
    mpi_interface : MPIInterface, optional
        In the case of MPI, create a more "lightweight" city district per MPI process only containing the particular
        buildings and assets that are required for the computations using MPI. Rank zero always has a list with all
        buildings.
    Returns
    ----------
    list of pycity_scheduling.classes.Building : list
        List of generated buildings.
    """
    if building_distribution is None:
        share = 1/len(tbd)
        building_distribution = {b: share for b in tbd}
    if heating_distribution is None:
        share = 1/len(heating_devices)
        heating_distribution = {d: share for d in heating_devices}
    if device_probabilities is None:
        device_probabilities = {'FL': 1}

    building_dicts = []
    for building, amount in _distribute(building_distribution, number).items():
        building_dicts += [tbd[building]] * amount
    assert len(building_dicts) == number

    heating_list = []
    for heating, amount in _distribute(heating_distribution, number).items():
        heating_list += [heating_devices[heating]] * amount
    assert len(heating_list) == number

    if any(map(lambda x: not 0 <= x <= 1, device_probabilities.values())):
        raise ValueError("Bad device probabilities")

    number_ap = sum(building['apartments'] for building in building_dicts)

    # apartment entities
    a = round(device_probabilities.get('FL', 0) * number_ap)
    fl_list = [True] * a + [False] * (number_ap - a)
    a = round(device_probabilities.get('DL', 0) * number_ap)
    dl_list = [True] * a + [False] * (number_ap - a)
    a = round(device_probabilities.get('EV', 0) * number_ap)
    ev_list = [True] * a + [False] * (number_ap - a)

    # building entities
    a = round(device_probabilities.get('PV', 0) * number)
    pv_list = [True] * a + [False] * (number - a)
    a = round(device_probabilities.get('BAT', 0) * number)
    bat_list = [True] * a + [False] * (number - a)

    dl_time_ranges = _calculate_dl_times(environment.timer)

    if seed is not None:
        random.seed(seed)
    random.shuffle(heating_list)
    random.shuffle(fl_list)
    random.shuffle(dl_list)
    random.shuffle(pv_list)
    random.shuffle(ev_list)
    random.shuffle(bat_list)

    buildings = []
    ap_counter = 0

    mpi_process_range = np.array([])
    if mpi_interface is not None:
        mpi_process_range = mpi_interface.get_mpi_process_range(number+1)

    # Generate buildings:
    for i, b in enumerate(building_dicts):
        dummy_building = False
        if mpi_interface is not None:
            if mpi_interface.get_rank() != 0:
                if mpi_interface.get_rank() != mpi_process_range[i+1]:
                    dummy_building = True

        building_type = b['building_type']
        name = 'BD{:03}_{}'.format(i + 1, building_type)

        bd = Building(environment, objective=objective, name=name,
                      profile_type=b['th_profile_type'],
                      building_type=building_type)

        bes = BuildingEnergySystem(environment)
        bd.addEntity(bes)

        ap_area = b['net_floor_area']/b['apartments']

        for n in range(b['apartments']):
            ap = Apartment(environment, ap_area)
            sh = SpaceHeating(environment, method=1, living_area=ap_area,
                              specific_demand=b['th_demand'],
                              profile_type=b['th_profile_type'])
            if not dummy_building:
                ap.addEntity(sh)

            if fl_list[ap_counter]:
                fl = FixedLoad(environment, method=1,
                               annual_demand=b['el_demand'],
                               profile_type=b['el_profile_type'])
                if not dummy_building:
                    ap.addEntity(fl)

            if dl_list[ap_counter]:
                e_el = random.uniform(0.8, 4.5)
                p_el = random.uniform(1.125, 2.5)
                time = random.choice(dl_time_ranges)
                dl = DeferrableLoad(environment, p_el_nom=p_el,
                                    e_consumption=e_el, load_time=time,
                                    lt_pattern='daily')
                if not dummy_building:
                    ap.addEntity(dl)

            if ev_list[ap_counter]:
                ev_data = random.choice(list(evd.values()))
                ev_charging_time = np.ones(environment.timer.simu_horizon)
                soc = random.uniform(0.2, 0.4)
                ev = ElectricVehicle(environment,
                                     e_el_max=ev_data['e_el_storage_max'],
                                     p_el_max_charge=11.0,
                                     soc_init=soc,
                                     charging_time=ev_charging_time,
                                     simulate_driving=False,
                                     minimum_soc_end=0.8,
                                     eta=0.95)
                if not dummy_building:
                    ap.addEntity(ev)

            if not dummy_building:
                bd.addEntity(ap)
            ap_counter += 1

        if not dummy_building:
            p_th_heat = max(sum(e.p_th_heat_schedule for e in filter_entities(bd, 'SH'))) + 1
        else:
            p_th_heat = 1000.0
        heating_device = heating_list[i](environment, p_th_nom=p_th_heat)
        ths = ThermalHeatingStorage(environment, e_th_max=2.0*p_th_heat, soc_init=0.5, loss_factor=10)

        if not dummy_building:
            bes.addDevice(heating_device)
            bes.addDevice(ths)

        if pv_list[i]:
            if b['roof_angle'] == 0.0:
                angle = 35.0
            else:
                angle = b['roof_angle']
            area = b['roof_area']/2.0
            # Solar world 290 standard values
            pv = Photovoltaic(environment, method=0, area=area, eta_noct=0.161853,
                              t_cell_noct=46, alpha_noct=-0.0041, beta=angle)
            if not dummy_building:
                bes.addDevice(pv)

        if bat_list[i]:
            p_el = 10.0 * b['apartments']
            bat = Battery(environment, e_el_max=p_el, p_el_max_charge=4.6,
                          p_el_max_discharge=4.6, soc_init=0.5, eta=0.95)
            if not dummy_building:
                bes.addDevice(bat)

        buildings.append(bd)
    return buildings


def generate_tabula_district(environment,
                             number_sfh,
                             number_mfh,
                             sfh_building_distribution=None,
                             sfh_heating_distribution=None,
                             sfh_device_probabilities=None,
                             mfh_building_distribution=None,
                             mfh_heating_distribution=None,
                             mfh_device_probabilities=None,
                             district_objective='price',
                             building_objective='price',
                             seed=1,
                             mpi_interface=None):
    """
    Generate a TABULA-based city district.

    Parameters
    ----------
    environment : pycity_scheduling.classes.Environment
    number_sfh : int
        Number of SFH buildings.
    number_mfh : int
        Number of MFH buildings.
    sfh_building_distribution : dict, optional
        The distribution of the houses among the tabula standard buildings. If
        omitted an equal distribution will be used.
        Keys : str
            'DE.N.<SFH|MFH>.<n>.Gen' or '<SFH|MFH>.<year>'
        Values : float
            Number between 0 and 1. The sum over all values must be one.
    sfh_heating_distribution : dict, optional
        The distribution of heating devices among the houses. If omitted an
        equal distribution will be used.
        Keys : str
            {'HP', 'EH', 'CHP', 'BL'}
        Values : float
            Number between 0 and 1. The sum over all values must be one.
    sfh_device_probabilities : dict, optional
        The probabilities of the houses / apartments to have the given device.
        Keys : str
            {'FL', 'DL', 'EV', 'PV', 'BAT'}
        Values : float
            Number between 0 and 1.
    mfh_building_distribution : dict, optional
        The distribution of the houses among the tabula standard buildings. If
        omitted an equal distribution will be used.
        Keys : str
            'DE.N.<SFH|MFH>.<n>.Gen' or '<SFH|MFH>.<year>'
        Values : float
            Number between 0 and 1. The sum over all values must be one.
    mfh_heating_distribution : dict, optional
        The distribution of heating devices among the houses. If omitted an
        equal distribution will be used.
        Keys : str
            {'HP', 'EH', 'CHP', 'BL'}
        Values : float
            Number between 0 and 1. The sum over all values must be one.
    mfh_device_probabilities : dict, optional
        The probabilities of the houses / apartments to have the given device.
        Keys : str
            {'FL', 'DL', 'EV', 'PV', 'BAT'}
        Values : float
            Number between 0 and 1.
    district_objective : str, optional
        Objective function for the city district (district operator). Default is 'price'.
    building_objective : str, optional
        Objective function for the buildings. Default is 'price'.
    seed: int, optional
        Specify a seed for the randomization. If omitted, a non-deterministic
        city district will be generated.
    mpi_interface : MPIInterface, optional
        In the case of MPI, create a more "lightweight" city district per MPI process only containing the particular
        buildings and assets that are required for the computations using MPI. Rank zero always has the city district
        with all buildings.
    Returns
    ----------
    list of pycity_scheduling.classes.CityDistrict : CityDistrict
        CityDistrict object containing the generated buildings.

    """
    cd = CityDistrict(environment, district_objective)
    # noinspection PyListCreation
    building_list = []
    building_list.extend(generate_tabula_buildings(environment,
                                                   number_sfh,
                                                   sfh_building_distribution,
                                                   sfh_heating_distribution,
                                                   sfh_device_probabilities,
                                                   building_objective,
                                                   seed,
                                                   mpi_interface
                                                   ))
    building_list.extend(generate_tabula_buildings(environment,
                                                   number_mfh,
                                                   mfh_building_distribution,
                                                   mfh_heating_distribution,
                                                   mfh_device_probabilities,
                                                   building_objective,
                                                   seed+1,
                                                   mpi_interface
                                                   ))
    positions = [[0, i+1] for i in range(len(building_list))]
    cd.addMultipleEntities(building_list, positions)
    return cd


def generate_simple_building(env, fl=0, sh=0, eh=0, ths=0, bat=0):
    """
    Generate a simple building with load and storage units.

    Parameters
    ----------
    env : pycity_scheduling.classes.Environment
    fl : float, optional
        Demand of the FixedLoad in [kW].
    sh : float, optional
        Demand of the SpaceHeating in [kW].
    eh : float, optional
        Power of the ElectricHeater in [kW].
    ths : float, optional
        Capacity of the ThermalHeatingStorage in [kWh].
    bat : float, optional
        Capacity of the Battery in [kWh].

    Returns
    -------
    pycity_scheduling.classes.Building: Building
        Simple building with load and storage units.
    """
    ti = env.timer

    bd = Building(env)
    ap = Apartment(env)
    bd.addEntity(ap)
    bes = BuildingEnergySystem(env)
    bd.addEntity(bes)
    if fl:
        ap.addEntity(FixedLoad(env, demand=np.full(ti.simu_horizon, fl)))
    if sh:
        ap.addEntity(SpaceHeating(env, loadcurve=np.full(ti.simu_horizon, sh)))
    if eh:
        bes.addDevice(ElectricHeater(env, p_th_nom=eh))
    if ths:
        bes.addDevice(ThermalHeatingStorage(env, e_th_max=ths, soc_init=0.5))
    if bat:
        bes.addDevice(Battery(env, e_el_max=bat, p_el_max_charge=bat/ti.time_slot))
    return bd



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import datetime
import calendar
import warnings
import pycity_base.classes.timer as ti


class Timer(ti.Timer):
    """
    Extension of pyCity_base class Timer for scheduling purposes

    Parameters
    ----------
    step_size : int, optional
        Number of seconds in one time step.
    op_horizon : int, optional
        Number of time steps used in one scheduling optimization.
    mpc_horizon : int, optional
        Number of time steps in whole simulation. All data must be
        available for this time. If `None` set value to `op_horizon`.
    mpc_step_width: int, optional
        Number of timesteps between two optimizations in MPC. If None set
        value to `op_horizon`.
    initial_date : tuple, optional
        Initial date in the format (year, month, day).
    initial_time : tuple, optional
        Initial time in the format (hour, minute, second).

    Notes
    -----
    - This class' behaviour may differ from the one of the baseclass, as it
      keeps an actual date rather than a relative counter only
    """

    def __init__(self, step_size=900, op_horizon=96,
                 mpc_horizon=None, mpc_step_width=None,
                 initial_date=(2015, 1, 1), initial_time=(0, 0, 0)):
        a, b = divmod(3600, step_size)
        if b != 0:
            raise ValueError(
                "`step_size` must be a factor of 3600 (1h)."
            )
        self._dt = datetime.datetime(
            initial_date[0], initial_date[1], initial_date[2],
            initial_time[0], initial_time[1], initial_time[2]
        )
        dt0 = datetime.datetime(self._dt.year, 1, 1)
        initial_day = dt0.weekday() + 1
        a, b = divmod((self._dt - dt0).total_seconds(), step_size)
        if b > 0:
            raise ValueError(
                "The difference from the starting point of the simulation "
                "to the beginning of the year must be a multiple of "
                "`time_discretization`."
            )
        # Workaround to make the weather instance calculate values for a whole
        # year.
        horizon = int(365 * 24 * 3600 / step_size)
        # if calendar.isleap(initial_date[0]):
        #     horizon += 24 * 3600 / step_size
        super(Timer, self).__init__(time_discretization=step_size,
                                    timesteps_used_horizon=op_horizon,
                                    timesteps_horizon=horizon,
                                    timesteps_total=horizon,
                                    initial_day=initial_day)

        self.time_slot = step_size / 3600
        self._init_dt = self._dt
        if mpc_horizon is None:
            self.simu_horizon = op_horizon
        else:
            self.simu_horizon = mpc_horizon
        if self.simu_horizon > horizon:
            horizon_name = "op_horizon" if mpc_horizon is None else "mpc_horizon"
            warnings.warn("`{}` indicates a horizon larger than one year. pyCity_base is currently not able to "
                          "produce forecasts for such a duration.".format(horizon_name), UserWarning)
        if mpc_step_width is None:
            self.mpc_step_width = op_horizon
        else:
            self.mpc_step_width = mpc_step_width

    @property
    def datetime(self):
        """
        The datetime of the current timestep.
        """
        return self._dt

    @property
    def date(self):
        """
        The date of the current timestep.
        """
        return datetime.date(self._dt.year, self._dt.month, self._dt.day)

    @property
    def is_leap(self):
        """
        If the year of the current timestep is a leap year.
        """
        return calendar.isleap(self._dt.year)

    @property
    def year(self):
        """
        The year of the current timestep.
        """
        return self._dt.year

    @property
    def month(self):
        """
        The month of the current timestep.
        """
        return self._dt.month

    @property
    def day(self):
        """
        The day of the month for the current timestep.
        """
        return self._dt.day

    @property
    def weekday(self):
        """
        The weekday of the current timestep.
        """
        return self._dt.weekday()

    @property
    def time(self):
        """
        The time for the current timestep.
        """
        return datetime.time(self._dt.hour, self._dt.minute, self._dt.second)

    @property
    def hour(self):
        """
        The hour of the current timestep.
        """
        return self._dt.hour

    @property
    def minute(self):
        """
        The minute of the current timestep.
        """
        return self._dt.minute

    @property
    def second(self):
        """
        The second of the current timestep.
        """
        return self._dt.second

    def print_datetime(self):
        """
        Print the datetime for the current timestep.
        """
        return str(self._dt)

    def mpc_update(self):
        """
        Update Timer for MPC.

        Move `self.mpc_step_width` timesteps forward.
        """
        self.current_timestep += self.mpc_step_width
        self._dt += datetime.timedelta(
            seconds=self.time_discretization*self.mpc_step_width
        )

        # update values from parent class
        self.current_weekday = self._dt.weekday() + 1
        self.current_day_weekend = self.current_weekday >= 6
        self.current_day = self.time_in_year("days")
        return

    def op_update(self):
        """
        Update Timer for a normal scheduling optimization.

        Go `self.timesteps_used_horizon` timesteps forward.
        """
        self.current_timestep += self.timesteps_used_horizon
        self._dt += datetime.timedelta(
            seconds=self.time_discretization*self.timesteps_used_horizon
        )

        # update values from parent class
        self.current_weekday = self._dt.weekday() + 1
        self.current_day_weekend = self.current_weekday >= 6
        self.current_day = self.time_in_year("days")
        return

    def reset(self):
        """
        Reset the Timer to the initial state.
        """
        self._dt = self._init_dt
        self.current_timestep = 0
        self.current_weekday = self._dt.weekday() + 1
        self.current_day_weekend = self.current_weekday >= 6
        self.current_day = self.time_in_year("days")
        return

    def time_in_year(self, unit="timesteps", from_init=False):
        """
        Time passed since the beginning of the year.

        Parameters
        ----------
        unit : str, optional
            Specifies the unit for the result.

            - 'timesteps' Return the time as timesteps.
            - 'seconds' Return the time as seconds.
            - 'days' Return the time as days.
        from_init : bool, optional
            Time for init date or current date.

        Returns
        -------
        int :
            Time in specified unit.
        """
        dt0 = datetime.datetime(self._dt.year, 1, 1)
        if from_init:
            dt1 = self._init_dt
        else:
            dt1 = self._dt
        seconds = (dt1 - dt0).total_seconds()
        if unit == "timesteps":
            a, b = divmod(seconds, self.time_discretization)
            if b > 0:
                raise ValueError(
                    "The difference from the starting point of the simulation "
                    "to the beginning of the year must be a multiple of "
                    "`time_discretization`."
                )
            return int(a)
        elif unit == "days":
            return seconds // 86400
        else:
            return seconds

    def time_in_week(self, unit="timesteps", from_init=False):
        """
        Time passed since beginning of the day.

        Parameters
        ----------
        unit : str, optional
            Specifies the unit for the result.

            - 'timesteps' Return the time as timesteps.
            - 'seconds' Return the time as seconds.
            - 'days' Return the time as days.
        from_init : bool, optional
            Time for init datetime or current datetime.

        Returns
        -------
        int :
            Time in specified unit.
        """
        if from_init:
            dt1 = self._init_dt
        else:
            dt1 = self._dt
        dt0 = datetime.datetime(dt1.year, dt1.month, dt1.day)
        dt0 -= datetime.timedelta(days=dt0.weekday())
        seconds = (dt1 - dt0).total_seconds()
        if unit == "timesteps":
            a, b = divmod(seconds, self.time_discretization)
            if b > 0:
                raise ValueError(
                    "The difference from the starting point of the simulation "
                    "to the beginning of the year must be a multiple of "
                    "`time_discretization`."
                )
            return int(a)
        elif unit == "days":
            return seconds // 86400
        else:
            return seconds

    def time_in_day(self, unit="timesteps", from_init=False):
        """
        Time passed since beginning of the day.

        Parameters
        ----------
        unit : str, optional
            Specifies the unit for the result.

            - 'timesteps' Return the time as timesteps.
            - 'seconds' Return the time as seconds.
            - 'days' Return the time as days.
        from_init : bool, optional
            Time for init datetime or current datetime.

        Returns
        -------
        int :
            Time in specified unit.
        """
        if from_init:
            dt1 = self._init_dt
        else:
            dt1 = self._dt
        dt0 = datetime.datetime(dt1.year, dt1.month, dt1.day)
        seconds = (dt1 - dt0).total_seconds()
        if unit == "timesteps":
            a, b = divmod(seconds, self.time_discretization)
            if b > 0:
                raise ValueError(
                    "The difference from the starting point of the simulation "
                    "to the beginning of the year must be a multiple of "
                    "`time_discretization`."
                )
            return int(a)
        elif unit == "days":
            return seconds // 86400
        else:
            return seconds



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np

from pycity_scheduling import constants, classes
from pycity_scheduling.classes import Boiler, ElectricalEntity, CityDistrict


def calculate_costs(entity, timestep=None, prices=None, feedin_factor=None):
    """
    Calculate electricity costs for the entity with the current schedule.

    Parameters
    ----------
    entity : ElectricalEntity
        The entity to calculate costs for.
    timestep : int, optional
        If specified, calculate metric only to this timestep.
    prices : array_like, optional
        Energy prices for simulation horizon.
    feedin_factor : float, optional
        Factor which is multiplied to the prices for feed-in revenue.

    Returns
    -------
    float :
        Electricity costs in [ct].
    """
    p = entity.p_el_schedule
    if prices is None:
        if isinstance(entity, CityDistrict):
            prices = entity.environment.prices.da_prices
        else:
            prices = entity.environment.prices.tou_prices
    if timestep:
        prices = prices[:timestep]
        p = p[:timestep]
    if feedin_factor is None:
        if isinstance(entity, CityDistrict):
            feedin_factor = 1
        else:
            feedin_factor = entity.environment.prices.feedin_factor
    costs = entity.time_slot * np.dot(prices[p > 0], p[p > 0])
    costs += entity.time_slot * np.dot(prices[p < 0], p[p < 0]) * feedin_factor
    return costs


def calculate_adj_costs(entity, schedule, timestep=None, prices=None, total_adjustments=True):
    """
    Calculate costs for power schedule adjustments.

    Parameters
    ----------
    entity : ElectricalEntity
        The entity to calculate adjustment costs for.
    schedule : str, optional
       Schedule to adjust to.

       - 'default' : Default schedule
       - 'ref' : Reference schedule
    timestep : int, optional
        If specified, calculate metric only to this timestep.
    prices : array_like, optional
        Adjustment prices for all timesteps in simulation horizon.
    total_adjustments : bool, optional
        `True` if positive and negative deviations shall be considered.
        `False` if only positive deviations shall be considered.

    Returns
    -------
    float :
        Adjustment costs in [ct].
    """
    if prices is None:
        if isinstance(entity, CityDistrict):
            prices = entity.environment.prices.da_prices
        else:
            prices = entity.environment.prices.tou_prices
    if timestep:
        prices = prices[:timestep]
    adj_power = calculate_adj_power(entity, schedule, timestep, total_adjustments)
    costs = entity.time_slot * np.dot(adj_power, prices)
    return costs


def calculate_co2(entity, timestep=None, co2_emissions=None):
    """
    Calculate CO2 emissions for the entity with the current schedule.

    Parameters
    ----------
    entity : OptimizationEntity
        The entity to calculate co2 emission for.
    timestep : int, optional
        If specified, calculate metric only to this timestep.
    co2_emissions : array_like, optional
        Specific CO2 emissions for all timesteps in the simulation horizon in [g/kWh].

    Returns
    -------
    float :
        CO2 emissions in [g].
    """
    if isinstance(entity, Boiler):
        p = entity.p_th_heat_schedule
        if timestep is not None:
            p = p[:timestep]
        co2 = -(sum(p) * entity.time_slot / entity.eta
                * constants.CO2_EMISSIONS_GAS)
        return co2
    elif isinstance(entity, ElectricalEntity):
        if timestep is None:
            timestep = len(entity.p_el_schedule)
        p = entity.p_el_schedule[:timestep]
        if co2_emissions is None:
            co2_emissions = entity.environment.prices.co2_prices
        co2_emissions = co2_emissions[:timestep]
        bat_schedule = sum(
            e.p_el_schedule[:timestep]
            for e in classes.filter_entities(entity, 'BAT')
        )
        # ToDo: This approach neglects battery losses.
        p = p - bat_schedule
        co2 = entity.time_slot * np.dot(p[p>0], co2_emissions[p>0])
        gas_schedule = sum(
            e.p_th_heat_schedule[:timestep].sum()
            * (1+e.sigma) / e.omega
            for e in classes.filter_entities(entity, 'CHP')
        )
        gas_schedule += sum(
            e.p_th_heat_schedule[:timestep].sum()
            / e.eta
            for e in classes.filter_entities(entity, 'BL')
        )
        pv_schedule = sum(
            e.p_el_schedule[:timestep].sum()
            for e in classes.filter_entities(entity, 'PV')
        )
        wec_schedule = sum(
            e.p_el_schedule[:timestep].sum()
            for e in classes.filter_entities(entity, 'WEC')
        )
        co2 -= gas_schedule * entity.time_slot * constants.CO2_EMISSIONS_GAS
        co2 -= pv_schedule * entity.time_slot * constants.CO2_EMISSIONS_PV
        co2 -= wec_schedule * entity.time_slot * constants.CO2_EMISSIONS_WIND
        return co2
    else:
        raise NotImplementedError


def calculate_adj_power(entity, schedule, timestep=None, total_adjustments=True):
    """
    Compute the power schedule adjustments.

    Parameters
    ----------
    entity : ElectricalEntity
        The entity to calculate the power adjustment for.
    schedule : str, optional
       Schedule to adjust to.

       - 'default' : Default schedule
       - 'ref' : Reference schedule
    timestep : int, optional
        If specified, calculate metric only to this timestep.
    total_adjustments : bool, optional
        `True` if positive and negative deviations shall be considered.
        `False` if only positive deviations shall be considered.

    Returns
    -------
    array of float :
        Adjustment power in [kW].
    """
    adjustments = entity.schedules[schedule]["p_el"] - entity.p_el_schedule
    if timestep:
        adjustments = adjustments[:timestep]
    if total_adjustments:
        return abs(adjustments)
    else:
        return np.maximum(adjustments, 0)


def calculate_adj_energy(entity, schedule, timestep=None, total_adjustments=True):
    """
    Compute the cumulative absolute energy of all adjustments.

    Parameters
    ----------
    entity : ElectricalEntity
        The entity to calculate the energy adjustment for.
    schedule : str, optional
       Schedule to adjust to.

       - 'default' : Default schedule
       - 'ref' : Reference schedule
    timestep : int, optional
        If specified, calculate metric only to this timestep.
    total_adjustments : bool, optional
        `True` if positive and negative deviations shall be considered.
        `False` if only positive deviations shall be considered.

    Returns
    -------
    float :
        Adjustments in [kWh].
    """
    p = calculate_adj_power(entity, schedule, timestep, total_adjustments)
    adjustments = entity.time_slot * sum(p)
    return adjustments


def metric_delta_g(entity, schedule):
    """
    Compute the factor ∆g for the current schedule and the reference schedule.

    Parameters
    ----------
    entity : ElectricalEntity
        The Entity to calculate the delta g metric for.
    schedule : str
       Referenced Schedule

       - 'default' : Normal schedule
       - 'ref' : Reference schedule

    Returns
    -------
    float :
        Factor ∆g.

    Notes
    -----
    - Implementation as given in the lecture "Elektrizitaetswirtschaft"
      by Prof. Dr.-Ing. Christian Rehtanz from TU Dortmund, Germany.
    """
    p_el_min_dsm = min(entity.p_el_schedule)
    p_el_max_dsm = max(entity.p_el_schedule)
    p_el_min_ref = min(entity.schedules[schedule]["p_el"])
    p_el_max_ref = max(entity.schedules[schedule]["p_el"])
    g = 1.0 - (abs(p_el_max_dsm - p_el_min_dsm) / abs(p_el_max_ref - p_el_min_ref))
    return g


def peak_to_average_ratio(entity, timestep=None):
    """
    Compute the ratio of peak demand to the average demand.

    The ratio of the absolute peak demand of the specified schedule is
    compared to the absolute mean of the schedule.
    It holds `r` >= 1, where a lower value is `better`, 1.0 would be optimal (i.e. no peaks at all).

    Parameters
    ----------
    entity : ElectricalEntity
        The entity to calculate the peak to avereage ratio for.
    timestep : int, optional
        If specified, calculate metric only to this timestep.

    Returns
    -------
    float :
        Peak to average ratio.
    """
    p = entity.p_el_schedule
    if timestep is not None:
        p = p[:timestep]
    peak = max(max(p), -min(p))
    mean = abs(np.mean(p))
    r = peak / mean
    return r


def peak_reduction_ratio(entity, schedule, timestep=None):
    """
    Compute the ratio of the peak reduction.

    The reduction of the absolute peak demand of the current schedule is
    compared to the peak demand in the reference schedule.
    If `r` < 1, the specified schedule has smaller peaks, otherwise the
    reference schedule has smaller peaks.
    Usually a small `r` value is the desired outcome.

    Parameters
    ----------
    entity : ElectricalEntity
        The entity to calculate the peak reduction ratio for.
    schedule : str
        Name of Schedule to compare to.

        - 'default' : Default schedule
        - 'ref' : Reference schedule
    timestep : int, optional
        If specified, calculate metric only to this timestep.

    Returns
    -------
    float :
        Peak reduction ratio.
    """
    if timestep is None:
        timestep = len(entity.p_el_schedule)
    p = entity.p_el_schedule[:timestep]
    ref = entity.schedules[schedule]["p_el"][:timestep]
    dr_peak = max(max(p), -min(p))
    ref_peak = max(max(ref), -min(ref))
    r = (dr_peak - ref_peak) / ref_peak
    return r


def self_consumption(entity, timestep=None):
    """
    Calculate the self-consumption rate for the current schedule.

    Parameters
    ----------
    entity : ElectricalEntity
        The entity to calculate the self-consumption rate for.
    timestep : int, optional
        If specified, calculate metric only to this timestep.

    Returns
    -------
    float :
        Self-consumption rate.
    """
    if timestep is None:
        timestep = len(entity.p_el_schedule)
    p = entity.p_el_schedule[:timestep]
    res_schedule = sum(e.p_el_schedule[:timestep] for e in classes.filter_entities(entity,
                                                                                   'generation_devices'))
    if not isinstance(res_schedule, np.ndarray):
        return 0
    generation = sum(res_schedule)
    if generation == 0:
        return 1
    neg_load = res_schedule - p
    np.clip(neg_load, a_min=None, a_max=0, out=neg_load)
    consumption = sum(np.maximum(neg_load, res_schedule))
    entity_consumption = consumption / generation
    return entity_consumption


def autarky(entity, timestep=None):
    """
    Calculate the autarky rate for the current schedule.

    Parameters
    ----------
    entity : ElectricalEntity
        The entity to calculate the autarky rate for.
    timestep : int, optional
        If specified, calculate metric only to this timestep.

    Returns
    -------
    float :
        Autarky rate.
    """
    if timestep is None:
        timestep = len(entity.p_el_schedule)
    p = entity.p_el_schedule[:timestep]
    res_schedule = - sum(e.p_el_schedule[:timestep] for e in classes.filter_entities(entity,
                                                                                     'generation_devices'))
    if not isinstance(res_schedule, np.ndarray):
        return 0
    load = p + res_schedule
    np.clip(load, a_min=0, a_max=None, out=load)
    consumption = sum(load)
    if consumption == 0:
        return 1
    cover = sum(np.minimum(res_schedule, load))
    autarky_val = cover / consumption
    return autarky_val


def absolute_flexibility_gain(entity, schedule, timestep=None):
    """
    Calculates the absolute flexibility gain for the entity with the current schedule.
    This corresponds to the amount of electrical energy shifted due to the selected objective function.

    Parameters
    ----------
    entity : ElectricalEntity
        The entity to calculate the absolute flexibility gain for.
    schedule : str
        Name of Schedule to compare to.

       - 'default' : Default schedule
       - 'ref' : Reference schedule
    timestep : int, optional
        If specified, calculate metric only to this timestep.

    Returns
    -------
    float :
        Absolute flexibility gain in [kWh].
    """
    if timestep is None:
        timestep = len(entity.p_el_schedule)
    p = entity.p_el_schedule[:timestep]
    ref = entity.schedules[schedule]["p_el"][:timestep]
    diff = ref - p
    np.clip(diff, a_min=0, a_max=None, out=diff)
    abs_flex = sum(diff) * entity.time_slot
    return abs_flex



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
import pycity_base.classes.demand.space_cooling as sc

from pycity_scheduling.classes.thermal_entity_cooling import ThermalEntityCooling


class SpaceCooling(ThermalEntityCooling, sc.SpaceCooling):
    """
    Extension of pyCity_base class SpaceCooling for scheduling purposes.

    As for all uncontrollable loads, the `p_th_schedule` contains the forecast
    of the load.

    Parameters
    ----------
    environment : Environment object
        Common to all other objects. Includes time and weather instances
    method : integer, optional
        - `0` : Provide load curve directly
        - `1` : Use thermal standard load profile (not implemented yet!)
    loadcurve : Array-like, optional
        Load curve for all investigated time steps
        Requires `method=0`
    living_area : Float, optional
        Living area of the apartment in m^2
        Requires `method=1`
    specific_demand : Float, optional
        Specific thermal demand of the building in kWh/(m^2 a)
        Requires `method=1`
    profile_type : str, optional
        Thermal SLP profile name
        Requires `method=1`
        - `HEF` : Single family household
        - `HMF` : Multi family household
        - `GBA` : Bakeries
        - `GBD` : Other services
        - `GBH` : Accomodations
        - `GGA` : Restaurants
        - `GGB` : Gardening
        - `GHA` : Retailers
        - `GHD` : Summed load profile business, trade and services
        - `GKO` : Banks, insurances, public institutions
        - `GMF` : Household similar businesses
        - `GMK` : Automotive
        - `GPD` : Paper and printing
        - `GWA` : Laundries

    Notes
    -----
    - The following constraint is added for removing the bounds from the TEC:

    .. math::
        p_{th\\_cool} = load\\_curve
    """

    def __init__(self, environment, method=0, loadcurve=1, living_area=0, specific_demand=0, profile_type='HEF'):

        super().__init__(environment, method, loadcurve*1000, living_area, specific_demand, profile_type)
        self._long_id = "SC_" + self._id_string

        ts = self.timer.time_in_year(from_init=True)
        p = self.loadcurve[ts:ts+self.simu_horizon] / 1000
        self.p_th_cool_schedule = p

    def populate_model(self, model, mode="convex"):
        super().populate_model(model, mode)
        m = self.model
        timestep = self.timestep

        def p_th_cool_vars_rule(model, t):
            return model.p_th_cool_vars[t] == self.p_th_cool_schedule[timestep + t]

        m.p_th_cool_vars_constr = pyomo.Constraint(m.t, rule=p_th_cool_vars_rule)
        return

    def update_model(self, mode=""):
        m = self.model
        timestep = self.timestep

        p_th_cool_vars_constr_active = m.p_th_cool_vars_constr.active

        def p_th_cool_vars_rule(model, t):
            return model.p_th_cool_vars[t] == self.p_th_cool_schedule[timestep + t]

        try:
            m.del_component(m.p_th_cool_vars_constr)
        except AttributeError:
            pass
        m.p_th_cool_vars_constr = pyomo.Constraint(m.t, rule=p_th_cool_vars_rule)

        if p_th_cool_vars_constr_active is False:
            m.p_th_cool_vars_constr.deactivate()
        return

    def new_schedule(self, schedule):
        super().new_schedule(schedule)
        self.copy_schedule(schedule, "default", "p_th_cool")
        return

    def update_schedule(self):
        pass

    def reset(self, schedule=None):
        pass



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
import pycity_base.classes.demand.electrical_demand as ed

from pycity_scheduling.classes.electrical_entity import ElectricalEntity


class FixedLoad(ElectricalEntity, ed.ElectricalDemand):
    """
    Extension of pyCity_base class ElectricalDemand for scheduling purposes.

    As for all uncontrollable loads, the `p_el_schedule` contains the forecast
    of the load.

    Parameters
    ----------
    environment : Environment
        Common Environment instance.
    method : int, optional
        Defaults to method zero.

        - 0 : provide load curve directly
        - 1 : standard load profile (for households)
        - 2 : stochastic electrical load model
        - 3 : annual profile based on measured weekly profiles (non-residential)
        - 4 : annual profile based on measured annual profiles (non-residential)
    demand : numpy.ndarray, optional
        Demand for all investigated time steps in [kW] when using `method=0`.
    annual_demand : float, optional
        Required for SLP and recommended for method 2.
        Defines the annual electrical demand in [kWh].
        If method 2 is chosen but no value is given, a standard value for
        Germany (https://lena.sachsen-anhalt.de/fileadmin/Bibliothek/Sonstige_Webprojekte/Lena/Pressemitteilungen/
        Stromspiegel/Stromspiegel2014_Medienblatt.pdf, accessed on 2020/09/28) is used.
    profile_type : str, optional
        - H0 : Household
        - L0 : Farms
        - L1 : Farms with breeding / cattle
        - L2 : Farms without cattle
        - G0 : Business (general)
        - G1 : Business (workingdays 8:00 AM - 6:00 PM)
        - G2 : Business with high loads in the evening
        - G3 : Business (24 hours)
        - G4 : Shops / Barbers
        - G5 : Bakery
        - G6 : Weekend operation
    total_nb_occupants : int, optional
        Number of people living in the household for `method=2`.
    randomize_appliances : bool, optional
        Randomization of appliances for `method=2`. Defaults to `True`.

        - True : distribute installed appliances randomly
        - False : use the standard distribution
    light_configuration : int, optional
        There are 100 light bulb configurations predefined for the
        stochastic model when using `method=2`. Defaults to zero.
        A value between 0 and 100 should be provided.
    occupancy : int, optional
        Occupancy given at 10-minute intervals for a full year for `method=2`.
    do_normalization : bool, optional
        Defines, if stochastic profile (`method=2`) should be
        normalized to given annual_demand value. Defaults to `False`.
        If set to `False`, annual el. demand depends on stochastic el. load
        profile generation. If set to `True`, does normalization with
        annual_demand.
    method_3_type : str, optional
        Defines type of profile for method=3. Defaults to `None`.
        Options:

        - 'food_pro': Food production
        - 'metal': Metal company
        - 'rest': Restaurant (with large cooling load)
        - 'sports': Sports hall
        - 'repair': Repair / metal shop
    method_4_type : str, optional
        Defines type of profile for method=4. Defaults to `None`.

        - 'metal_1' : Metal company with smooth profile
        - 'metal_2' : Metal company with fluctuation in profile
        - 'warehouse' : Warehouse
    prev_heat_dev : bool, optional
        Defines, if heating devices should be prevented within chosen
        appliances for `method=2`. Defaults to `False`.
        If set to True, DESWH, E-INST, Electric shower, Storage heaters
        and Other electric space heating are set to zero.
    app_filename : str, optional
        Path to Appliances file for `method=2`. Defaults to `None`.
        If set to None, uses default file Appliances.csv in
        /inputs/stochastic_electrical_load/.
    light_filename : str, optional
        Path to Lighting configuration file for `method=2`. Defaults to `None`.
        If set to None, uses default file Appliances.csv in
        /inputs/stochastic_electrical_load/.
    season_light_mod : bool, optional
        Defines, if cosine-wave should be used to strengthen seasonal
        influence on lighting. Defaults to `False`.
        If True, enlarges lighting power demand in winter month and reduces
        lighting power demand in summer month.
    light_mod_fac : float, optional
        Define factor, related to maximal lighting power, which is used
        to implement seasonal influence. Defaults to 25%.
        Only relevant, if `season_light_mod` == True

    Notes
    -----
    - Standard load profiles, for instance for Germany, can be found here:
      https://www.bdew.de/energie/standardlastprofile-strom/ (accessed on 2020/09/28)

    - Average German electricity consumption data per household can be found here:
      https://lena.sachsen-anhalt.de/fileadmin/Bibliothek/Sonstige_Webprojekte/Lena/Pressemitteilungen/
      Stromspiegel/Stromspiegel2014_Medienblatt.pdf (accessed on 2020/09/28)

    - The following constraint is added for removing the bounds from EE:

    .. math::
        p_{el} = load\\_curve
    """

    def __init__(self, environment, method=0, demand=0, annual_demand=0, profile_type="H0", single_family_house=True,
                 total_nb_occupants=0, randomize_appliances=True, light_configuration=0, occupancy=None,
                 do_normalization=False, method_3_type=None, method_4_type=None, prev_heat_dev=False, app_filename=None,
                 light_filename=None, season_light_mod=False, light_mod_fac=0.25):
        super().__init__(environment, method, demand*1000, annual_demand, profile_type, single_family_house,
                         total_nb_occupants, randomize_appliances, light_configuration, occupancy, do_normalization,
                         method_3_type, method_4_type, prev_heat_dev, app_filename, light_filename, season_light_mod,
                         light_mod_fac)
        self._long_id = "FL_" + self._id_string

        ts = self.timer.time_in_year(from_init=True)
        p = self.loadcurve[ts:ts+self.simu_horizon] / 1000
        self.p_el_schedule = p

    def populate_model(self, model, mode="convex"):
        super().populate_model(model, mode)
        m = self.model
        timestep = self.timestep

        def p_el_vars_rule(model, t):
            return model.p_el_vars[t] == self.p_el_schedule[timestep + t]
        m.p_el_vars_constr = pyomo.Constraint(m.t, rule=p_el_vars_rule)
        return

    def update_model(self, mode=""):
        m = self.model
        timestep = self.timestep

        p_el_vars_constr_active = m.p_el_vars_constr.active

        def p_el_vars_rule(model, t):
            return model.p_el_vars[t] == self.p_el_schedule[timestep + t]
        try:
            m.del_component(m.p_el_vars_constr)
        except AttributeError:
            pass
        m.p_el_vars_constr = pyomo.Constraint(m.t, rule=p_el_vars_rule)

        if p_el_vars_constr_active is False:
            m.p_el_vars_constr.deactivate()
        return

    def new_schedule(self, schedule):
        super().new_schedule(schedule)
        self.copy_schedule(schedule, "default", "p_el")
        return

    def update_schedule(self):
        pass

    def reset(self, schedule=None):
        pass



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
import pycity_base.classes.building as bd

from pycity_scheduling import classes
from pycity_scheduling.classes.entity_container import EntityContainer


class Building(EntityContainer, bd.Building):
    """
    Extension of pyCity_base class Building for scheduling purposes.

    Parameters
    ----------
    environment : Environment
        Common to all other objects. Includes time and weather instances.
    objective : str, optional
        Objective for the scheduling. The default is 'price'.

        - 'price' : Optimize for the prices given by `prices.tou_prices`.
        - 'co2' : Optimize for the CO2 emissions given by `prices.co2_prices`.
        - 'peak-shaving' : Try to flatten the schedule as much as possible.
        - 'max-consumption' : Try to reduce the maximum of the absolute values of the schedule as much as possible.
        - 'self-consumption' : Try to maximize the self-consumption of the local power generation.
        - 'least-squares': Try to `follow` a given reference power profile.
        - 'none' : No objective (leave all flexibility to other participants).
    name : str, optional
        Name for the building.
        If name is None, set it to self._long_id.
    profile_type : str, optional
        Thermal SLP profile name
        Requires `method=1`

        - 'HEF' : Single family household
        - 'HMF' : Multi family household
        - 'GBA' : Bakeries
        - 'GBD' : Other services
        - 'GBH' : Accommodations
        - 'GGA' : Restaurants
        - 'GGB' : Gardening
        - 'GHA' : Retailers
        - 'GHD' : Summed load profile business, trade and services
        - 'GKO' : Banks, insurances, public institutions
        - 'GMF' : Household similar businesses
        - 'GMK' : Automotive
        - 'GPD' : Paper and printing
        - 'GWA' : Laundries
    building_type : str, optional
        Build year profile name, the detailed list is implemented in
        `tabula_data.py`.
    storage_end_equality : bool, optional
        `True` if the soc at the end of the scheduling has to be equal to
        the initial soc.
        `False` if it has to be greater or equal than the initial soc.
    least_squares_profile : numpy.ndarray, optional
        Profile to be used by the least-squares objective.

    Notes
    -----
    - The exchange of thermal energy between different buildings is currently not supported.
      As a result, the building additionally adds the following set of constrains to the
      ones of the EntityContainer:

    .. math::
        p_{th\\_heat} &=& 0 \\\\
        p_{th\\_cool} &=& 0

    - The building can also add robustness constrains for thermal heating storage:

    .. math::
        e_{u\\_bound} \\geq \\sum_i e_{th\\_heat\\_i} \\geq e_{l\\_bound} \\\\

    - The :math:`E_{u\\_bound}` and :math:`E_{l\\_bound}` are determined by the
      robustness parameter, the available capacity of thermal heating storage, the magnitude of heating
      required by SpaceHeating and the magnitude of heating that can be produced by the building's heating units.
    """

    def __init__(self, environment, objective='price', name=None,
                 profile_type=None, building_type=None,
                 storage_end_equality=False, least_squares_profile=None):
        super().__init__(environment)

        self._long_id = "BD_" + self._id_string
        if name is None:
            self.name = self._long_id
        else:
            self.name = name

        self.objective = objective
        self.profile_type = profile_type
        self.building_type = building_type
        self.storage_end_equality = storage_end_equality
        self.least_squares_profile = least_squares_profile

    def populate_model(self, model, mode="convex", robustness=None):
        """
        Add building block to pyomo ConcreteModel.

        Call parent's `populate_model` method and set variables lower
        bounds to `None`. Then call `populate_model` method of the BES
        and all contained apartments and add constraints that the sum
        of their variables for each period equals the corresponding
        own variable.

        Parameters
        ----------
        model : pyomo.ConcreteModel
        mode : str, optional
            Specifies which set of constraints to use.

            - `convex`  : Use linear constraints
            - `integer`  : Use same constraints as convex mode
        robustness : tuple, optional
            Tuple of two floats. First entry defines how many time steps are
            protected from deviations. Second entry defines the magnitude of
            deviations which are considered.
        """
        if not self.has_bes:
            raise AttributeError(
                "No BES object in %s, but a BES is always required.\nModeling aborted." % str(self)
            )
        super().populate_model(model, mode)
        m = self.model

        def p_th_cool_equality_rule(model, t):
            return model.p_th_cool_vars[t] == 0
        m.p_th_cool_equality_constr = pyomo.Constraint(m.t, rule=p_th_cool_equality_rule)

        def p_th_heat_equality_rule(model, t):
            return model.p_th_heat_vars[t] == 0
        m.p_th_heat_equality_constr = pyomo.Constraint(m.t, rule=p_th_heat_equality_rule)

        if robustness is not None and (self.bes.getHasDevices(all_devices=False, ths=True)[0]):
            self._create_robust_constraints()
        return

    def update_model(self, mode="", robustness=None):
        """
        Update block parameters and bounds.

        Set parameters and bounds according to the current situation of the device
        according to the previous schedule and the current forecasts.

        Parameters
        ----------
        mode : str, optional
            Specifies which set of constraints to use.

            - `convex`  : Use linear constraints
            - `integer`  : Use same constraints as convex mode
        robustness : tuple, optional
            Tuple of two floats. First entry defines how many time steps are
            protected from deviations. Second entry defines the magnitude of
            deviations which are considered.
        """
        super().update_model(mode)

        if robustness is not None and (self.bes.getHasDevices(all_devices=False, ths=True)[0]):
            self._update_robust_constraints(robustness)
        return

    def _create_robust_constraints(self):
        # ToDo: Robust constraints currently available for heating devices only. Also include them for cooling devices.

        m = self.model
        ths_ms = [ths.model for ths in self.bes.ths_units]
        m.lower_robustness_bounds = pyomo.Param(m.t, mutable=True)
        m.upper_robustness_bounds = pyomo.Param(m.t, mutable=True)

        def e_lower_rule(model, t):
            return sum(ths.e_th_heat_vars[t] for ths in ths_ms) >= model.lower_robustness_bounds[t]
        m.lower_robustness_constr = pyomo.Constraint(m.t, rule=e_lower_rule)

        def e_upper_rule(model, t):
            return sum(ths.e_th_heat_vars[t] for ths in ths_ms) <= model.upper_robustness_bounds[t]
        m.upper_robustness_constr = pyomo.Constraint(m.t, rule=e_upper_rule)
        return

    def _update_robust_constraints(self, robustness):
        # ToDo: Robust constraints currently available for heating devices only. Also include them for cooling devices.

        m = self.model
        timestep = self.timer.current_timestep
        e_th_max = sum(ths.e_th_max for ths in self.bes.ths_units)
        end_value = sum(ths.soc_init * ths.e_th_max for ths in self.bes.ths_units)
        uncertain_p_th = np.zeros(self.op_horizon)
        p_th_demand_sum = np.zeros(self.op_horizon)

        # aggregate demand from all thermal heating demands
        t1 = timestep
        t2 = timestep + self.op_horizon
        for apartment in self.apartments:
            for entity in apartment.th_heating_demand_list:
                if isinstance(entity, classes.SpaceHeating):
                    p_th_demand_sum += entity.p_th_heat_schedule[t1:t2]

        p_th_max_supply = sum(
            e.p_th_nom for e in classes.filter_entities(self, 'heating_devices')
        )

        # Get parameter for robustness conservativeness
        lambda_i, lambda_d = divmod(robustness[0], 1)
        lambda_i = int(lambda_i)
        if lambda_i >= self.op_horizon:
            lambda_i = self.op_horizon - 1
            lambda_d = 1

        for t in self.op_time_vec:
            uncertain_p_th[t] = min(p_th_demand_sum[t] * robustness[1],
                                    p_th_max_supply)
            tmp = sorted(uncertain_p_th, reverse=True)
            uncertain_e_th = self.time_slot * (sum(tmp[:lambda_i])
                                               + tmp[lambda_i]*lambda_d)

            # If environment is too uncertain, keep storage on half load
            if uncertain_e_th >= e_th_max / 2:
                m.lower_robustness_bounds[t] = e_th_max / 2
                m.upper_robustness_bounds[t] = e_th_max / 2
            # standard case in
            elif t < self.op_horizon - 1:
                m.upper_robustness_bounds[t] = e_th_max - uncertain_e_th
                m.lower_robustness_bounds[t] = uncertain_e_th
            # set storage to uncertain_e_th or set SOC_End to soc_init to
            # prevent depletion of storage
            else:
                if e_th_max / end_value <= 0.5:
                    end_value = max(end_value, uncertain_e_th)
                else:
                    end_value = min(end_value, e_th_max - uncertain_e_th)
                m.lower_robustness_bounds[t] = end_value
                m.upper_robustness_bounds[t] = end_value
        return

    def get_lower_entities(self):
        if self.has_bes:
            yield self.bes
        yield from self.apartments



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import matplotlib.pyplot as plt

import pycity_scheduling.util.factory as factory
import pycity_scheduling.util.debug as debug
import pycity_scheduling.util.mpi_interface as mpi
from pycity_scheduling.util.metric import self_consumption, autarky
from pycity_scheduling.algorithms import *
from pycity_scheduling.solvers import *


# In this example, the power schedule for a city district scenario is determined by means of real parallel distributed
# optimization algorithm implementations using the Message Passing Interface (MPI) standard. Use command "mpiexec" to
# create a defined number of parallel MPI processes.
# The scenario is built upon the district setup as defined in example 'example_17_district_generator.py',
# but it contains just a few buildings for the sake of demonstration.

def main(do_plot=False):
    # First, instantiate the MPI interface and disable printing for all MPI processes other than with rank 0.
    mpi_interface = mpi.MPIInterface()
    mpi_interface.disable_multiple_printing(stdout=True, stderr=True)

    print("\n\n------ Example 09: Algorithm Parallel MPI ------\n\n")

    # Create an environment using the factory's "generate_standard_environment" method. The environment
    # automatically encapsulates time, weather, and price data/information.
    env = factory.generate_standard_environment(initial_date=(2018, 10, 6), step_size=3600, op_horizon=24)

    # Create single-family houses:
    num_sfh = 3

    # 50% SFH.2002, 30% SFH.2010, 20% SFH.2016 (based on TABULA):
    sfh_distribution = {
        'SFH.2002': 0.5,
        'SFH.2010': 0.3,
        'SFH.2016': 0.2,
    }

    # 50% of the single-family houses are equipped with heat pump, 10% with boiler, and 40% with electric heater:
    sfh_heating_distribution = {
        'HP': 0.5,
        'BL': 0.1,
        'EH': 0.4,
    }

    # All single-family houses are equipped with a fixed load, 0% have a deferrable load, and 30% have an electric
    # vehicle. Moreover, 50% of all single-family houses have a battery unit and 100% have a rooftop photovoltaic unit
    # installation.
    # The values are rounded in case they cannot be perfectly matched to the given number of buildings.
    sfh_device_probs = {
        'FL': 1.0,
        'DL': 0.0,
        'EV': 0.3,
        'BAT': 0.5,
        'PV': 1.0,
    }

    # Create multi-family houses (number of apartments according to TABULA):
    num_mfh = 0

    # 60% MFH.2002, 20% SFH.2010, 20% SFH.2016 (based on TABULA):
    mfh_distribution = {
        'MFH.2002': 0.6,
        'MFH.2010': 0.2,
        'MFH.2016': 0.2,
    }

    # 40% of the multi-family houses are equipped with heat pump, 20% with boiler, and 40% with electric heater:
    mfh_heating_distribution = {
        'HP': 0.4,
        'BL': 0.2,
        'EH': 0.4,
    }

    # All apartments inside a multi-family houses are equipped with a fixed load, 0% have a deferrable load, and 20%
    # have an electric vehicle. Moreover, 40% of all multi-family houses have a battery unit and 100% have a rooftop
    # photovoltaic unit installation.
    # The values are rounded in case they cannot be perfectly matched to the given number of buildings.
    mfh_device_probs = {
        'FL': 1.0,
        'DL': 0.0,
        'EV': 0.2,
        'BAT': 0.4,
        'PV': 1.0,
    }

    # Finally, create the desired city district using the factory's "generate_tabula_district" method. The district's
    # district operator's objective is defined as "peak-shaving" and the buildings' objectives are defined as
    # "peak-shaving", too.
    district = factory.generate_tabula_district(env, num_sfh, num_mfh,
                                                sfh_distribution,
                                                sfh_heating_distribution,
                                                sfh_device_probs,
                                                mfh_distribution,
                                                mfh_heating_distribution,
                                                mfh_device_probs,
                                                district_objective='peak-shaving',
                                                building_objective='peak-shaving',
                                                mpi_interface=mpi_interface
                                                )

    # Hierarchically print the district and all buildings/assets:
    debug.print_district(district, 2)

    # Perform the city district scheduling using the central optimization algorithm without integer constraints as a
    # reference:
    print("\n### Central Algorithm (Convex) ###\n")
    opt = CentralOptimization(district, mode="convex")
    results = opt.solve()
    district.copy_schedule("central-convex")

    # Print the building's schedules and some metrics:
    for building in district.get_lower_entities():
        print("Schedule building {}:".format(str(building)))
        print(list(building.p_el_schedule))
    print("Schedule of the city district:")
    print(list(district.p_el_schedule))
    print("")
    print("Self-consumption rate: {: >4.2f}".format(self_consumption(district)))
    print("Autarky rate: {: >4.2f}".format(autarky(district)))
    print("Objective_value: ", results["obj_value"])

    # Perform the city district scheduling using the central optimization algorithm with integer constraints as a
    # reference:
    print("\n### Central Algorithm (Integer) ###\n")
    opt = CentralOptimization(district, mode="integer")
    results = opt.solve()
    district.copy_schedule("central-integer")

    # Print the building's schedules and some metrics:
    for building in district.get_lower_entities():
        print("Schedule building {}:".format(str(building)))
        print(list(building.p_el_schedule))
    print("Schedule of the city district:")
    print(list(district.p_el_schedule))
    print("")
    print("Self-consumption rate: {: >4.2f}".format(self_consumption(district)))
    print("Autarky rate: {: >4.2f}".format(autarky(district)))
    print("Objective_value: ", results["obj_value"])

    # Perform the city district scheduling using the Dual Decomposition optimization algorithm without integer
    # constraints:
    print("\n### Dual Decomposition Algorithm MPI (Convex) ###\n")
    opt = DualDecompositionMPI(district, mpi_interface, rho=0.1, eps_primal=0.1, max_iterations=200)
    results = opt.solve()
    district.copy_schedule("dual-decomposition")

    # Print the building's schedules:
    for building in district.get_lower_entities():
        print("Schedule building {}:".format(str(building)))
        print(list(building.p_el_schedule))
    print("Schedule of the city district:")
    print(list(district.p_el_schedule))
    print("")
    print("Self-consumption rate: {: >4.2f}".format(self_consumption(district)))
    print("Autarky rate: {: >4.2f}".format(autarky(district)))
    print("Iterations: ", results["iterations"][-1])
    print("Objective_value: ", results["obj_value"][-1])

    # Perform the city district scheduling using the Exchange ADMM optimization algorithm without integer constraints:
    print("\n### Exchange ADMM Algorithm MPI (Convex) ###\n")
    opt = ExchangeADMMMPI(district, mpi_interface, rho=2.0, eps_primal=0.1, eps_dual=1.0, max_iterations=200)
    results = opt.solve()
    district.copy_schedule("exchange-admm")

    # Print the building's schedules:
    for building in district.get_lower_entities():
        print("Schedule building {}:".format(str(building)))
        print(list(building.p_el_schedule))
    print("Schedule of the city district:")
    print(list(district.p_el_schedule))
    print("")
    print("Self-consumption rate: {: >4.2f}".format(self_consumption(district)))
    print("Autarky rate: {: >4.2f}".format(autarky(district)))
    print("Iterations: ", results["iterations"][-1])
    print("Objective_value: ", results["obj_value"][-1])

    # Perform the city district scheduling using the Exchange MIQP ADMM optimization algorithm (constrained) with
    # integer constraints:
    print("\n### Constrained Exchange MIQP ADMM Algorithm MPI (Integer) ###\n")
    opt = ExchangeMIQPADMMMPI(district, mpi_interface, mode='integer', x_update_mode='constrained',
                              eps_exch_primal=0.1, eps_exch_dual=1.0, gamma=100.0, gamma_incr=1.01, rho=10.0,
                              max_iterations=200, varying_penalty_parameter=False)
    results = opt.solve()
    district.copy_schedule("exchange_miqp_admm-constrained")

    # Print the building's schedules:
    for building in district.get_lower_entities():
        print("Schedule building {}:".format(str(building)))
        print(list(building.p_el_schedule))
    print("Schedule of the city district:")
    print(list(district.p_el_schedule))
    print("")
    print("Self-consumption rate: {: >4.2f}".format(self_consumption(district)))
    print("Autarky rate: {: >4.2f}".format(autarky(district)))
    print("Iterations: ", results["iterations"][-1])
    print("Objective_value: ", results["obj_value"][-1])

    return


# Conclusions:
# Using MPI, we can facilitate the real parallel execution of distributed optimization algorithms - even on
# loosely-coupled machines. The parallel implementations allow for significant computational speedups and are therefore
# suitable to tackle large-scale optimization problems that may not be solved by an off-the-shelf centralized
# computation because of scalability issues.


if __name__ == '__main__':
    # Run example:
    main(do_plot=True)



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
from pycity_scheduling.classes.electrical_entity import ElectricalEntity
from pycity_scheduling.util import approximation as approx


class ElectricalLine(ElectricalEntity):
    """
    Class ElectricalLine for scheduling purposes. Represents a connecting line/cable for the interconnection of
    buildings in pycity_scheduling.

    Parameters
    ----------
    environment : pycity_scheduling.classes.Environment
        Common to all other objects. Includes time and weather instances.
    sending_node : Union[ElectricalNode, Building]
        First entity to connect, sending node of the line.
    receiving_node : Union[ElectricalNode, Building]
        Second entity to connect, receiving node of the line.
    resistance : float, optional
        Line resistance in Ohm.
    reactance : float, optional
        Line reactance in Ohm.
    max_capacity : float, optional
        Maximum line capacity (power) in kVA.
    """
    def __init__(self, environment, sending_node, receiving_node, resistance, reactance, max_capacity, *args, **kwargs):
        super().__init__(environment, *args, **kwargs)
        self._kind = "electricalline"
        self._long_id = "EL_" + self._id_string
        self.sending_node = sending_node
        self.receiving_node = receiving_node
        self.resistance = resistance
        self.reactance = reactance
        self.max_capacity = max_capacity

        # Power flow through line:
        self.new_var("p_line_el_vars")
        self.new_var("q_line_el_vars")

        # Approximated square of power flow:
        self.new_var("p_app_el_vars")
        self.new_var("q_app_el_vars")

        # Power losses through line:
        self.new_var("p_loss_el_vars")
        self.new_var("q_loss_el_vars")

        # Binary variables:
        self.new_var("x_p")
        self.new_var("x_q")

    def populate_model(self, model, mode='convex'):
        """
        Add electrical line block to pyomo ConcreteModel.

        Call parent's `populate_model` method.
        Set bounds on power flow according to line capacity.
        Perform max affine approximation for the square function. Coefficients of the approximation are used to set
        lower bounds on power square `p_app_el_vars`.
        Set disjunctive upper bounds constraints on `p_app_el_vars` using the big-M method.
        Set equality constraints to calculate power losses based on square power approximation.
        Add voltage drop constraint.
        
        Parameters
        ----------
        model : pyomo.ConcreteModel
        mode : str, optional
            Specifies which set of constraints to use.

            - `convex`  : Use linear constraints
            - `integer`  : Use same constraints as convex mode
        """
        super().populate_model(model, mode)
        m = self.model

        m.q_el_vars = pyomo.Var(m.t, domain=pyomo.Reals, initialize=0)

        # Introduce power flow through the line as an additional variable for readability
        m.p_line_el_vars = pyomo.Var(m.t, domain=pyomo.Reals, initialize=1)

        def p_line_el_vars_rule(model, t):
            return m.p_line_el_vars[t] == m.p_el_vars[t]
        m.p_line_el_vars_constr = pyomo.Constraint(m.t, rule=p_line_el_vars_rule)

        # Power flow can be positive or negative considering the arbitrary orientation of the line:
        m.p_el_vars.setlb(None)

        # Perform max-affine approximation of power square:
        square_function = lambda x: x**2
        inf_bound = [-np.max(self.max_capacity)]
        sup_bound = [np.max(self.max_capacity)]
        nb_samples = 1000
        nb_segments = 15

        # Calculate coefficients of affine functions:
        slopes, intercepts = approx.piecewise_linear_approx(square_function, inf_bound, sup_bound, nb_samples,
                                                            nb_segments)
        slopes = np.array(slopes, dtype=np.float32)
        intercepts = np.array(intercepts, dtype=np.float32)

        # Set constraints for max-affine approximation of power square:
        M = 1000  # for big-M method
        max_affine_index = pyomo.RangeSet(0, nb_segments-1)
        m.p_app_el_vars = pyomo.Var(m.t, domain=pyomo.Reals, bounds=(0, None), initialize=0)   # power square
        m.q_app_el_vars = pyomo.Var(m.t, domain=pyomo.Reals, bounds=(0, None), initialize=0)

        # x_p[i][t]=1 if p[t] is in the ith approximation segment, 0 otherwise:
        m.x_p = pyomo.Var(max_affine_index, m.t, domain=pyomo.Binary)
        m.x_q = pyomo.Var(max_affine_index, m.t, domain=pyomo.Binary)

        def p_lower_bounds_approx_rule(model, t, i):
            return m.p_app_el_vars[t] >= slopes[i] * m.p_el_vars[t] + intercepts[i]
        m.p_lower_bounds_approx_constr = pyomo.Constraint(m.t, max_affine_index,
                                                          rule=p_lower_bounds_approx_rule)

        def q_lower_bounds_approx_rule(model, t, i):
            return m.q_app_el_vars[t] >= slopes[i] * m.q_el_vars[t] + intercepts[i]
        m.q_lower_bounds_approx_constr = pyomo.Constraint(m.t, max_affine_index,
                                                          rule=q_lower_bounds_approx_rule)

        # Big-M method to set disjunctive upper bound constraints and handle saturation of lower bounds constraints:
        def p_upper_bounds_approx_rule(model, t, i):
            return m.p_app_el_vars[t] - (slopes[i] * m.p_el_vars[t] + intercepts[i]) <= M * (1-m.x_p[i, t])
        m.p_upper_bounds_approx_constr = pyomo.Constraint(m.t, max_affine_index,
                                                          rule=p_upper_bounds_approx_rule)
        
        def q_upper_bounds_approx_rule(model, t, i):
            return m.q_app_el_vars[t] - (slopes[i] * m.q_el_vars[t] + intercepts[i]) <= M * (1-m.x_q[i, t])
        m.q_upper_bounds_approx_constr = pyomo.Constraint(m.t, max_affine_index,
                                                          rule=q_upper_bounds_approx_rule)

        # At least one lower bound approx constraint is saturated (the corresponding upper bound constraint is active):
        def p_saturation_rule(model, t):
            return sum(m.x_p[i, t] for i in range(nb_segments)) >= 1
        m.p_saturation_constr = pyomo.Constraint(m.t, rule=p_saturation_rule)

        def q_saturation_rule(model, t):
            return sum(m.x_q[i, t] for i in range(nb_segments)) >= 1
        m.q_saturation_constr = pyomo.Constraint(m.t, rule=q_saturation_rule)

        # Power loss is positive and calculated from the max-affine approximation:
        m.p_loss_el_vars = pyomo.Var(m.t, domain=pyomo.Reals, bounds=(0, None), initialize=0)
        m.q_loss_el_vars = pyomo.Var(m.t, domain=pyomo.Reals, bounds=(0, None), initialize=0)

        def p_loss_rule(model, t):
            return m.p_loss_el_vars[t] == (m.p_app_el_vars[t] + m.q_app_el_vars[t]) * self.resistance
        m.p_loss_constr = pyomo.Constraint(m.t, rule=p_loss_rule)

        def q_loss_rule(model, t):
            return m.q_loss_el_vars[t] == (m.p_app_el_vars[t] + m.q_app_el_vars[t]) * self.reactance
        m.q_loss_constr = pyomo.Constraint(m.t, rule=q_loss_rule)

        # Bounded power flow for healthy operation
        def bounded_power_flow_rule(model, t):
            return m.p_app_el_vars[t] + m.q_app_el_vars[t] <= self.max_capacity**2
        m.bounded_power_flow_constr = pyomo.Constraint(m.t, rule=bounded_power_flow_rule)

        # Voltage drop across the line, calculated from power flow and power loss through the line:
        def voltage_drop_rule(model, t):
            return (self.sending_node.model.V_el_vars[t] - self.receiving_node.model.V_el_vars[t] ==
                    2 * self.resistance*m.p_el_vars[t] + 2 * self.reactance*m.q_el_vars[t] -
                    self.resistance*m.p_loss_el_vars[t] - self.reactance*m.q_loss_el_vars[t])
        m.voltage_drop_constr = pyomo.Constraint(m.t, rule=voltage_drop_rule)



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.cluster import KMeans


def piecewise_linear_approx(f, inf_bound, sup_bound, nb_samples=1000, nb_segments=10):
    """
    Perform piece-wise linear approximation for convex or concave 1D functions.

    Find a partition of the bound interval using KMeans-clustering with x_sample and function slope as features, for a
    given number of clusters. Then perform a linear regression on each segment of the partition.
    The approximation is given by the maximum (resp. minimum) of all linear functions if the function to approximate is
    convex (resp. concave). The convexity of the function is given through the argument "convexity" that takes the value
    "convex" if the function is convex and anything else if the function is concave.
    
    Parameters
    ----------
    f : callable
        Function to be approximated.
    inf_bound : list of float
        List of lower bounds for the variables in var.
    sup_bound : list of float
        List of upper bounds for the variables in var.
    nb_samples : int, default 100
        Number of sampling points.
    nb_segments : int, default 10
        Number of segments used in the partitioning of the bound interval.

    Returns
    -------
    slopes : List[float]
    intercepts : List[float]
    """
    # Sampling:
    x_sample = np.linspace(inf_bound, sup_bound, nb_samples)
    y_sample = f(x_sample)

    # Calculating gradient:
    gradient_list = []
    for i in range(1, len(x_sample) - 1):
        delta_x = x_sample[i + 1] - x_sample[i - 1]
        delta_y = y_sample[i + 1] - y_sample[i - 1]
        gradient = delta_y / delta_x
        gradient_list.append(gradient)
    gradient_list.insert(0, (y_sample[1] - y_sample[0]) / (x_sample[1] - x_sample[0]))
    gradient_list.append((y_sample[-1] - y_sample[-2]) / (x_sample[-1] - x_sample[-2]))

    # Building features such as position and gradient:
    x_sample = x_sample.reshape(-1, 1)
    gradient_list = np.array(gradient_list).reshape(-1, 1)
    features = np.column_stack((x_sample, gradient_list))
    mean = np.mean(features, axis=0)
    std_dev = np.std(features, axis=0)
    normalized_features = (features - mean) / std_dev

    # Perform KMeans clustering:
    kmeans = KMeans(n_clusters=nb_segments, n_init=10, random_state=0)
    clusters = kmeans.fit_predict(normalized_features)

    # Linear interpolation on each cluster:
    slope_list = []
    intercept_list = []
    for i in range(nb_segments):
        mask = (clusters == i)
        x_segment = x_sample[mask]
        y_segment = y_sample[mask]
        model = LinearRegression()
        model.fit(x_segment, y_segment)
        slope = model.coef_[0]
        intercept = model.intercept_
        slope_list.append(slope)
        intercept_list.append(intercept)

    return slope_list, intercept_list



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
import pycity_base.classes.supply.compression_chiller as cc

from pycity_scheduling.classes.thermal_entity_cooling import ThermalEntityCooling
from pycity_scheduling.classes.electrical_entity import ElectricalEntity
from pycity_scheduling.util.generic_constraints import LowerActivationLimit


class Chiller(ThermalEntityCooling, ElectricalEntity, cc.CompressionChiller):
    """
    Extension of pyCity_base class CompressionChiller for scheduling purposes.

    Parameters
    ----------
    environment : Environment
        Common to all other objects. Includes time and weather instances.
    p_th_nom : float
        Nominal thermal power of the chiller in [kW].
    cop : numpy.ndarray or int or float, optional
        The chiller's coefficient of performance (without unit)
        If array, it must provide the coefficient of performance (cop) for
        each time step in the simulation horizon.
        If int or float, a constant cop over the whole horizon is assumed.
        If omitted, an air-water chiller is assumed and the cop is
        calculated based on the ambient air temperature, eta and t_flow.
    eta : int or float, optional
        The Carnot efficiency of the chiller.
        If cop is omitted, eta is used to calculate the cop based on T_sink and T_source
        according to: cop = eta * T_sink * (T_sink - T_source) with T_sink and T_source in Kelvin
    t_min : float, optional
        The chiller's minimum provided temperature in °C
    lower_activation_limit : float, optional (only adhered to in integer mode)
        Must be in [0, 1]. Lower activation limit of the chiller as a
        percentage of the rated power. When the chiller is in operation, its
        power must be zero or between the lower activation limit and its
        rated power.

        - `lower_activation_limit = 0`: Linear behavior
        - `lower_activation_limit = 1`: Two-point controlled
    min_off_time : int, optional (only adhered to in integer mode)
        Minimum number of consecutive time steps the chiller must remain turned off, once it has been decided that it
        should turn off subject to 'lower_activation_limit'. Defaults to '1', i.e., not limited to a consecutive number
        of time steps.
    min_on_time : int, optional (only adhered to in integer mode)
        Minimum number of consecutive time steps the chiller must remain turned on, once it has been decided that it
        should turn on subject to 'lower_activation_limit'. Defaults to '1', i.e., not limited to a consecutive number
        of time steps.
    t_flow: float, optional
        T_sink temperature delivered by CC in °C.
        Is used for cop calculation if cop is omitted.
        Note that t_flow >= t_min must hold.

    Notes
    -----
    - CHs offer sets of constraints for operation. In the `convex` mode the
      following constraints and bounds are generated by the CH:

    .. math::
        0 \\geq p_{th\\_cool} &\\geq& -p_{th\\_nom} \\\\
        cop * p_{el} &=& - p_{th\\_cool}

    - See also:
        - pycity_scheduling.util.generic_constraints.LowerActivationLimit:
          Generates additional constraints for the `lower_activation_limit` in `integer` mode.
    """

    def __init__(self, environment, p_th_nom, cop=None, eta=0.36, t_min=4.0, lower_activation_limit=0, min_off_time=1,
                 min_on_time=1, t_flow=16.0):
        simu_horizon = environment.timer.simu_horizon

        (t_ambient,) = environment.weather.getWeatherForecast(getTAmbient=True)
        ts = environment.timer.time_in_year()
        t_ambient = t_ambient[ts:ts + simu_horizon]
        if cop is None:
            # ToDo: Variable cop calculation could be improved for chillers and it is recommended to use a fixed cop.
            with np.errstate(divide='ignore', invalid='ignore'):
                cop = -eta * np.true_divide((t_flow + 273.15), (t_flow - t_ambient))
                cop = np.nan_to_num(cop)
                cop[cop < 0] = 0
            print(list(cop))
        elif isinstance(cop, (int, float)):
            cop = np.full(simu_horizon, cop)
        elif not isinstance(cop, np.ndarray):
            raise TypeError(
                "Unknown type for `cop`: {}. Must be `numpy.ndarray`, `int` "
                "or `float`".format(type(cop))
            )
        super().__init__(environment, p_th_nom*1000.0, cop, t_min, lower_activation_limit)
        self._long_id = "CH_" + self._id_string
        self.cop = cop
        self.p_th_nom = p_th_nom
        self.min_off_time = min_off_time
        self.min_on_time = min_on_time
        self.activation_constr = None

    def populate_model(self, model, mode="convex"):
        """
        Add device block to pyomo ConcreteModel.

        Call parent's `populate_model` method and set thermal variables lower
        bounds to `-self.p_th_nom` and the upper bounds to zero. Also add
        constraint to bind electrical demand to thermal output.

        Parameters
        ----------
        model : pyomo.ConcreteModel
        mode : str, optional
            Specifies which set of constraints to use

            - `convex`  : Use linear constraints
            - `integer`  : Use integer variables representing discrete control decisions
        """
        if self.activation_constr is None:
            self.activation_constr = LowerActivationLimit(self, "p_th_cool", self.lower_activation_limit,
                                                          -self.p_th_nom, self.min_off_time, self.min_on_time)

        super().populate_model(model, mode)
        m = self.model

        if mode == "convex" or "integer":
            def p_th_cool_vars_lb_rule(model, t):
                return model.p_th_cool_vars[t] >= -self.p_th_nom
            m.p_th_cool_vars_lb_constr = pyomo.Constraint(m.t, rule=p_th_cool_vars_lb_rule)

            def p_th_cool_vars_ub_rule(model, t):
                return model.p_th_cool_vars[t] <= 0.0
            m.p_th_cool_vars_ub_constr = pyomo.Constraint(m.t, rule=p_th_cool_vars_ub_rule)

            m.cop = pyomo.Param(m.t, mutable=True)

            def p_coupl_rule(model, t):
                return model.p_th_cool_vars[t] + model.cop[t] * model.p_el_vars[t] == 0
            m.p_coupl_constr = pyomo.Constraint(m.t, rule=p_coupl_rule)

            self.activation_constr.apply(m, mode)
        else:
            raise ValueError(
                "Mode %s is not implemented by class Chiller." % str(mode)
            )
        return

    def update_model(self, mode=""):
        m = self.model
        cop = self.cop[self.op_slice]
        for t in self.op_time_vec:
            m.cop[t] = cop[t]
        return



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import csv
import json


__all__ = [
    'schedule_to_dict',
    'schedule_to_csv',
    'schedule_to_json',
]


def schedule_to_dict(input_list, schedule=None):
    """
    Create a dictionary containing a reference to the optimized schedules.

    Parameters
    ----------
    input_list : list
        List of entities.
    schedule : str or list, optional
           Schedule or list of schedules to save.

           - `None` : Current schedule of entity
           - 'default' : Normal schedule
           - 'ref' : Reference schedule
    """
    if isinstance(schedule, str) or schedule is None:
        schedules = [schedule]
    elif isinstance(schedule, list):
        schedules = schedule
    else:
        raise ValueError("Unknown type for schedule argument.")
    entities = {str(entity): {} for entity in input_list}
    for ent, ent_schedules in zip(input_list, entities.values()):
        for schedule in schedules:
            if schedule is None:
                schedule = ent.current_schedule_active

            sub_schedules = ent.schedules[schedule].copy()
            ent_schedules[schedule] = sub_schedules
    return entities


def schedule_to_json(input_list, file_name, schedule=None):
    """
    Write the optimized schedule of all entities to a json file.

    Parameters
    ----------
    input_list : list
        List of entities.
    file_name : str or file-like object
        Specify the file name or an open file where the json should be saved
        in. If file_name is a string and it does not have the `.json`
        extension it will be appended.
    schedule : str or list, optional
           Schedule or list of schedules to save.

           - `None` : Current schedule of entity
           - 'default' : Normal schedule
           - 'ref' : Reference schedule
    """
    entities = schedule_to_dict(input_list, schedule)
    for ent_schedules in entities.values():
        for sub_schedule in ent_schedules.values():
            for var_name in sub_schedule.keys():
                sub_schedule[var_name] = sub_schedule[var_name].tolist()

    if isinstance(file_name, str):
        if not file_name.endswith(".json"):
            file_name = file_name + ".json"
        with open(file_name, 'w') as outfile:
            json.dump(entities, outfile)
    else:
        json.dump(entities, file_name)


def schedule_to_csv(input_list, file_name, delimiter=";", schedule=None):
    """
    Write the optimized schedule of all entities to a CSV file.

    Parameters
    ----------
    input_list : list
        List of entities.
    file_name : str or file-like object
        Specify the file name or an open file where the csv should be saved
        in. If file_name is a string and it does not have the `.csv`
        extension it will be appended.
    delimiter : str
        CSV file delimiter character.
    schedule : str or list, optional
           Schedule or list of schedules to save.

           - `None` : Current schedule of entity
           - 'default' : Normal schedule
           - 'ref' : Reference schedule
    """
    sub_schedules = []
    headers = []

    max_schedule_length = 0
    entities = schedule_to_dict(input_list, schedule)
    for ent, ent_schedules in entities.items():
        for ent_schedule_name, sub_schedule in ent_schedules.items():
            for var_name, var_schedule in sub_schedule.items():
                headers.append("_".join([ent, ent_schedule_name, var_name]))
                sub_schedules.append(var_schedule+0.0)
                max_schedule_length = max(max_schedule_length, len(var_schedule))

    if isinstance(file_name, str) and not file_name.endswith(".csv"):
        file_name = file_name + ".csv"
    with open(str(file_name), 'w', newline='') as file:
        writer = csv.writer(file, delimiter=delimiter, quoting=csv.QUOTE_NONE)
        writer.writerow(headers)

        row = 0
        while row < max_schedule_length:
            row_string = []
            for schedule in sub_schedules:
                if row < len(schedule):
                    row_string.append(str(schedule[row]))
                else:
                    row_string.append('')
            writer.writerow(row_string)
            row += 1
    return



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import pyomo.environ as pyomo
import pycity_base.classes.city_district as cd

from pycity_scheduling.classes.electrical_entity import ElectricalEntity
from pycity_scheduling.classes.electrical_node import ElectricalNode
from pycity_scheduling.classes.electrical_grid import ElectricalGrid


class CityDistrict(ElectricalEntity, cd.CityDistrict):
    """
    Extension of pyCity_base class CityDistrict for scheduling purposes. Also represents the district operator.

    Parameters
    ----------
    environment : Environment
    objective : str, optional
        Objective for the district operator. Default is 'price'.

        - 'price' : Optimize for the minimum total cost given by `prices.da_prices`.
        - 'peak-shaving' : Try to 'flatten' the schedule as much as possible.
        - 'co2' : Optimize for the minimum total co2 emissions given by `prices.co2_prices`.
        - 'valley-filling' : Try to fill the 'valleys' given by `self.valley_profile`.
        - 'max-consumption' : Try to minimize the maximum power subscription.
        - 'self-consumption' : Try to maximize the self-consumption of the local power generation.
        - 'least-squares': Try to `follow` a given reference power profile.
        - 'flexibility-quantification' : To be used to quantify the flexibility potential of the city district only.
        - 'none' : No objective.
    valley_profile : numpy.ndarray, optional
        Profile to be filled by applying the valley filling objective.
    least_squares_profile : numpy.ndarray, optional
        Profile to be used by the least-squares objective.
    electrical_grid: ElectricalGrid, optional
        Assign an ElectricalGrid object to the city district representing the local power grid

    Notes
    -----
    - The constraints generated by a CD are the same as the ones created by an EntityContainer.
    """

    def __init__(self, environment, objective='price', valley_profile=None, least_squares_profile=None,
                 electrical_grid=None):
        super().__init__(environment)
        self._long_id = "CD_" + self._id_string
        self.objective = objective
        self.valley_profile = valley_profile
        self.least_squares_profile = least_squares_profile
        self.electrical_grid = electrical_grid

    def populate_model(self, model, mode="convex"):
        """
        Add city district block to pyomo ConcreteModel.

        Call parent's `populate_model` methods and set variables lower
        bounds to `None`.

        Parameters
        ----------
        model : pyomo.ConcreteModel
        mode : str, optional
            Specifies which set of constraints to use.

            - `convex`  : Use linear constraints
            - `integer`  : Use same constraints as convex mode
        """
        super().populate_model(model, mode)
        return

    def get_objective(self, coeff=1):
        if self.objective == 'valley-filling':
            obj = 0
            valley = self.valley_profile[self.op_slice]
            for t in range(self.op_horizon):
                obj += coeff * self.model.p_el_vars[t] * self.model.p_el_vars[t]
                obj += 2 * coeff * valley[t] * self.model.p_el_vars[t]
            return obj
        elif self.objective == 'price':
            prices = self.environment.prices.da_prices[self.op_slice]
            s = sum(abs(prices))
            if s > 0:
                obj = 0
                prices = prices * self.op_horizon / s
                for t in range(self.op_horizon):
                    obj += prices[t] * self.model.p_el_vars[t]
                return obj
            else:
                return 0
        else:
            return super().get_objective(coeff)

    def get_decision_var(self):
        if self.objective in ['price', 'valley-filling']:
            return self.model.p_el_vars
        else:
            return super().get_decision_var()

    def get_lower_entities(self):
        for node in self.nodes.values():
            yield node['entity']

    def account_imbalance(self):
        """
        Changes the current schedule to account imbalances.

        The imbalances are determined from the current schedules of the contained entities.
        """

        total = np.zeros_like(self.schedule["p_el"])
        for e in self.get_lower_entities():
            total += e.schedule["p_el"]

        self.schedule["p_el"] = total
        return

    def addElectricalGrid(self, environment, max_p_slack, max_q_slack, slack_V=None, ref_V = 400, V_min=0.95,
                          V_max=1.05):
        """
        Add electrical grid to the city district.

        Create an `ElectricalGrid` object and link it to the city district through the `electrical_grid` attribute.
        Add an electrical node the electrical grid for each entity contained in the city district.
        
        Parameters
        ----------
        environment : Environment
        max_p_slack : float
            Maximum active power that can be exchanged through slack node of the grid in kW.
        max_q_slack : float
            Maximum reactive power that can be exchanged through slack node of the grid in kVar.
        slack_V : float
            If not `None`, voltage at which the slack node should be operated in V. Default is None.
        ref_V : float
            Nominal voltage in the electrical grid in V. Default is 400V.
        V_min : float
            Minimal per unit voltage allowed in the electrical grid. Default is 0.95.
        V_max : float
            Maximal per unit voltage allowed in the electrical grid. Default is 1.05.
        """
        self.electrical_grid = ElectricalGrid(environment, self, max_p_slack, max_q_slack, slack_V, ref_V, V_min, V_max)
        for entity in self.get_lower_entities():
            related_node = ElectricalNode(environment, entity, None)
            self.electrical_grid.add_node(related_node)



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np

from pycity_scheduling.algorithms import *
from pycity_scheduling.solvers import *
import pycity_scheduling.util.factory as factory


# In this example, the power schedule for a complex city district scenario is determined. The scenario is built upon the
# district setup as defined in example 'example_16_district_generator.py', but it contains only 4 buildings for the
# sake of simplicity. Scheduling is done using the ALADIN optimization algorithm.


def main(do_plot=False):
    print("\n\n------ Example 19: Scheduling City District ALADIN ------\n\n")
    # First, create an environment using the factory's "generate_standard_environment" method. The environment
    # automatically encapsulates time, weather, and price data/information.
    env = factory.generate_standard_environment(initial_date=(2018, 1, 10), step_size=900, op_horizon=96)

    # Create N single-family houses:
    num_sfh = 4

    # TABULA building stock distribution:
    sfh_distribution = {
        'SFH.1200': 0.033,
        'SFH.1860': 0.096,
        'SFH.1919': 0.112,
        'SFH.1949': 0.085,
        'SFH.1958': 0.150,
        'SFH.1969': 0.145,
        'SFH.1979': 0.070,
        'SFH.1984': 0.11,
        'SFH.1995': 0.102,
        'SFH.2002': 0.077,
        'SFH.2010': 0.01,
        'SFH.2016': 0.01,
    }

    # All households are equipped with electro-thermal heat pump units:
    sfh_heating_distribution = {
        'HP': 1.0,
        'BL': 0.0,
        'EH': 0.0,
        'CHP': 0.0,
    }

    # All households are equipped with a fixed load and PV unit.
    # Moreover, 50% of all households have a stationary battery storage unit and 30% have an electric vehicle.
    # The values are rounded in case they cannot be perfectly matched to the given number of buildings.
    sfh_device_probs = {
        'FL': 1.0,
        'DL': 0.0,
        'EV': 0.3,
        'BAT': 0.5,
        'PV': 1.0,
    }

    # Finally, create the desired city district using the factory's "generate_tabula_district" method. The district
    # operator's objective is defined as "peak-shaving" and the buildings' objectives are defined as
    # "peak-shaving", too. Both objectives are quadratic and therefore the optimization problem becomes an MIQP.
    district = factory.generate_tabula_district(environment=env,
                                                number_sfh=num_sfh,
                                                number_mfh=0,
                                                sfh_building_distribution=sfh_distribution,
                                                sfh_heating_distribution=sfh_heating_distribution,
                                                sfh_device_probabilities=sfh_device_probs,
                                                district_objective='peak-shaving',
                                                building_objective='peak-shaving',
                                                seed=0
                                                )

    # Set the parameters for the ALADIN algorithm:
    max_iterations = 100
    mode = 'convex'

    if DEFAULT_SOLVER == "gurobi_direct" or DEFAULT_SOLVER == "gurobi_persistent":
        # Perform the city district scheduling using the central benchmark algorithm:
        opt = CentralOptimization(city_district=district, mode=mode)
        results = opt.solve()

        print("Central optimization completed.")

        # Print schedules:
        print("\nSystem level schedule:")
        print(list(district.p_el_schedule))
        print("\nLocal level schedules:")
        tmp = np.zeros_like(district.p_el_schedule)
        for bd in district.get_lower_entities():
            tmp += bd.p_el_schedule
            print(str(bd) + ":", list(bd.p_el_schedule))

        print("\nSum local level building schedules:")
        print(list(tmp))

        # Calculate and print the objective value:
        print("\nObjective value: ", results["obj_value"])
        print("\n\n")

        # Perform the city district scheduling using the ALADIN algorithm:
        opt = DerivativeFreeALADIN(city_district=district, mode=mode, max_iterations=max_iterations)
        results = opt.solve()

        print("Day-ahead ALADIN optimization completed.")

        # Print schedules:
        print("\nSystem level schedule:")
        print(list(district.p_el_schedule))
        print("\nLocal level schedules:")
        tmp = np.zeros_like(district.p_el_schedule)
        for bd in district.get_lower_entities():
            tmp += bd.p_el_schedule
            print(str(bd) + ":", list(bd.p_el_schedule))

        print("\nSum local level building schedules:")
        print(list(tmp))

        # Calculate and print the objective value:
        print("\nObjective value: ", results["obj_value"][-1])
    else:
        print("Algorithm ALADIN is currently supported by the gurobi solver only! Example is not executed.")
    return


if __name__ == '__main__':
    # Run the case study:
    main(do_plot=True)



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import matplotlib.pyplot as plt

from pycity_scheduling.classes import *
from pycity_scheduling.algorithms import *
from pycity_scheduling.util.plot_schedules import plot_entity, plot_imbalance, plot_entity_directory
from pycity_scheduling.util.write_schedules import schedule_to_json, schedule_to_csv


# This examples demonstrates some post-processing capabilities of the pycity_scheduling framework.


def main(do_plot=False):
    print("\n\n------ Example 29: Post-Processing Schedule Evaluation ------\n\n")

    # Define timer, price, weather, and environment objects:
    t = Timer()
    p = Prices(timer=t)
    w = Weather(timer=t)
    e = Environment(timer=t, weather=w, prices=p)

    # City district with district operator objective "peak-shaving":
    cd = CityDistrict(environment=e, objective='peak-shaving')

    # Schedule two sample buildings. The buildings' objectives are defined as "peak-shaving".

    # Building no. one comes with fixed load, space heating, electric heater, pv unit, thermal energy storage, and
    # electrical battery energy storage:
    bd1 = Building(environment=e, objective='peak-shaving')
    cd.addEntity(entity=bd1, position=[0, 0])
    bes = BuildingEnergySystem(environment=e)
    bd1.addEntity(bes)
    ths = ThermalHeatingStorage(environment=e, e_th_max=40, soc_init=0.5)
    bes.addDevice(ths)
    eh = ElectricHeater(environment=e, p_th_nom=10)
    bes.addDevice(eh)
    ap = Apartment(environment=e)
    bd1.addEntity(ap)
    load = np.array(-np.sin(np.linspace(0, 2.1*np.pi, 96)+1))
    fi = FixedLoad(e, method=0, demand=load)
    ap.addEntity(fi)
    sh = SpaceHeating(environment=e, method=0, loadcurve=load)
    ap.addEntity(sh)
    pv = Photovoltaic(environment=e, method=1, peak_power=4.6)
    bes.addDevice(pv)
    bat = Battery(environment=e, e_el_max=4.8, p_el_max_charge=3.6, p_el_max_discharge=3.6, eta=1.0)
    bes.addDevice(bat)

    # Building no. two comes with deferrable load, curtailable load, space heating, chp unit, thermal energy storage and
    # an electric vehicle:
    bd2 = Building(environment=e, objective='peak-shaving')
    cd.addEntity(entity=bd2, position=[0, 1])
    bes = BuildingEnergySystem(environment=e)
    bd2.addEntity(bes)
    ths = ThermalHeatingStorage(environment=e, e_th_max=35, soc_init=0.5)
    bes.addDevice(ths)
    chp = CombinedHeatPower(environment=e, p_th_nom=20.0)
    bes.addDevice(chp)
    ap = Apartment(environment=e)
    bd2.addEntity(ap)
    load = np.array(-2*np.sin(np.linspace(0, 1.9*np.pi, 96)+3))
    dl = DeferrableLoad(environment=e, p_el_nom=2.0, e_consumption=2.0)
    ap.addEntity(dl)
    cl = CurtailableLoad(environment=e, p_el_nom=1.6, max_curtailment=0.8)
    ap.addEntity(cl)
    sh = SpaceHeating(environment=e, method=0, loadcurve=load)
    ap.addEntity(sh)
    ev = ElectricVehicle(environment=e, e_el_max=37.0, p_el_max_charge=22.0, soc_init=0.65, simulate_driving=True,
                         minimum_soc_end=0.8)
    ap.addEntity(ev)

    # Perform the local scheduling:
    opt = LocalOptimization(city_district=cd)
    opt.solve()
    cd.copy_schedule("ref")

    # Perform the actual scheduling:
    opt = CentralOptimization(city_district=cd)
    opt.solve(beta=0.01)
    cd.copy_schedule("central")

    if not do_plot:
        fig, ax = plt.subplots()
        kwargs = {'ax': ax}
    else:
        kwargs = {}

    # As the power generation of a PV unit cannot be altered by the algorithms when force_renewables is not set to True,
    # the schedule is the same for 'ref' and 'default'.
    if do_plot:
        plot_entity(pv, schedule=["ref", "default"], **kwargs)

    # When the city district objective is not considered in the scheduling (= uncoordinated reference case), the
    # electricity schedule of the city district is usually not balanced. This is not the case for the coordinated case
    # ("default")
    if do_plot:
        plot_entity(cd, schedule=["ref", "default"], **kwargs)

    # In addition to this detail in the plots, you can notice that all the buildings have a constant p_th_heat_schedule
    # of zero. This is the result of all buildings satisfying a net thermal demand of zero.
    if do_plot:
        plot_entity(bd1, schedule=["ref", "default"], **kwargs)
        plot_entity(bd2, schedule=["ref", "default"], **kwargs)

    # The CHP unit has the optional integer variable p_th_Heat_state. Even when the integer mode is not used as in this
    # example, the schedule for this variable is still provided. For this variable, the value is always rounded to zero
    # if it is close enough to zero.
    if do_plot:
        plot_entity(chp, schedule=["ref", "default"], **kwargs)

    # The electric vehicle and battery objects also have an integer variable. This variable indicates if the battery
    # is either charging or discharging/not charging.
    if do_plot:
        plot_entity(ev, schedule=["ref", "default"], **kwargs)

    # The imbalance for the city district schedule in respect to the two buildings' schedules is rather large for the
    # default schedule. This is because of the relatively large eps_primal value for the ADMM algorithm in this example.
    if do_plot:
        plot_imbalance(cd, schedule=["ref", "default"], **kwargs)

    # Since a building is always solved in one solver instance for the ADMM algorithm, its individual imbalance is
    # small, though. A solver with a large primal feasibility tolerance could, however, yield larger imbalances.
    if do_plot:
        plot_imbalance(bd1, schedule=["ref", "default"], **kwargs)

    # If required, the schedules can also be saved as .json or .csv files:
    save_schedule = do_plot

    if save_schedule:
        entities = list(cd.get_all_entities())
        schedule_to_csv(entities, file_name="example_post-processing_schedule_evaluation", schedule=["ref", "default"])
        schedule_to_json(entities, file_name="example_post-processing_schedule_evaluation", schedule=["ref", "default"])

        # Plot the city district hierarchy and its lower-level devices into the current directory:
        plot_entity_directory(cd, ["ref", "default"], levels=2)
    return


# Conclusions:
# Useful post-processing functions, such as the plotting or exporting of power schedules can be used easily in
# framework pycity_scheduling.


if __name__ == '__main__':
    # Run example:
    main(do_plot=True)



"""
The pycity_scheduling framework


Copyright (C) 2025,
Institute for Automation of Complex Power Systems (ACS),
E.ON Energy Research Center (E.ON ERC),
RWTH Aachen University

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""


import numpy as np
import unittest
import tempfile
import filecmp
import pytest
import os
import os.path as op
import pyomo.environ as pyomo
from pyomo.solvers.plugins.solvers.persistent_solver import PersistentSolver

import pycity_scheduling.util as util
import pycity_scheduling.util.factory
import pycity_scheduling.util.generic_constraints as gen_constrs
from pycity_scheduling import solvers
from pycity_scheduling.classes import *
from pycity_scheduling.data.tabula_data import tabula_building_data
from pycity_scheduling.exceptions import SchedulingError
from pycity_scheduling.util.write_schedules import schedule_to_dict, schedule_to_csv, schedule_to_json
from pycity_scheduling.util.plot_schedules import plot_entity_directory
from pycity_scheduling.algorithms import CentralOptimization


def _get_constr_count(block):
    return sum(len(data) for data in block.component_map(pyomo.Constraint).itervalues())


class TestConstraints(unittest.TestCase):
    def setUp(self):
        t = Timer(step_size=3600, op_horizon=12)
        p = Prices(t)
        w = Weather(t)
        self.env = Environment(t, w, p)
        return

    def test_lower_activation_limit(self):
        ee = ElectricalEntity(self.env)
        m = pyomo.ConcreteModel()
        ee.populate_model(m, "integer")
        o = 0
        for t in range(len(ee.model.p_el_vars)):
            o += ee.model.p_el_vars[t] * ee.model.p_el_vars[t]
        o -= 2 * 0.3 * sum(ee.model.p_el_vars[t] for t in range(len(ee.model.p_el_vars)))
        m.o = pyomo.Objective(expr=o)
        self.assertEqual(_get_constr_count(ee.model), 12*2)
        lal_constr = gen_constrs.LowerActivationLimit(ee, "p_el", 0.5, 1)

        lal_constr.apply(ee.model, "convex")
        self.assertEqual(_get_constr_count(ee.model), 12*2)
        solve_model(m)
        ee.update_schedule()
        for i in ee.op_time_vec:
            self.assertAlmostEqual(ee.p_el_schedule[i], 0.3, 3)
            self.assertAlmostEqual(ee.p_el_state_schedule[i], 1.0, 3)

        lal_constr.apply(ee.model, "integer")
        self.assertEqual(_get_constr_count(ee.model), 12*4)

        solve_model(m)
        ee.update_schedule()
        for i in ee.op_time_vec:
            self.assertAlmostEqual(ee.p_el_schedule[i], 0.5, 3)
            self.assertAlmostEqual(ee.p_el_state_schedule[i], 1.0, 3)
        return


class TestFactory(unittest.TestCase):
    def setUp(self):
        t = Timer()
        p = Prices(t)
        w = Weather(t)
        self.env = Environment(t, w, p)
        self.sd = {'SFH.2002': 0.43, 'SFH.2010': 0.47, 'SFH.2016': 0.1}
        self.hd = {'HP': 0.31, 'BL': 0.34, 'EH': 0.35}
        self.dd = {'FL': 1, 'DL': 0.2, 'EV': 0.3, 'BAT': 0.5, 'PV': 0.8}
        self.md = {'MFH.2002': 0.99, 'MFH.2010': 0.01, 'MFH.2016': 0.0}
        return

    def test_distribution(self):
        district = util.factory.generate_tabula_district(self.env, 5, 2, self.sd, self.hd, self.dd,
                                                         self.md, self.hd, self.dd)
        buildings = [b for b in district.get_lower_entities() if type(b) == Building]
        for i, building in enumerate(self.sd.keys()):
            amount = [2, 2, 1][i]
            self.assertEqual(amount, sum(1 for b in buildings if b.building_type ==
                                         tabula_building_data[building]['building_type']))
        for i, building in enumerate(self.md.keys()):
            amount = [2, 0, 0][i]
            self.assertEqual(amount, sum(1 for b in buildings if b.building_type ==
                                         tabula_building_data[building]['building_type']))
        for i, h_id in enumerate(self.hd.keys()):
            amount = [1, 3, 3][i]
            self.assertEqual(amount, sum(1 for b in buildings for e in b.get_entities() if
                                         type(e) == heating_devices[h_id]))
        sfhs = [b for b in buildings if b.building_type.find("SFH") != -1]
        mfhs = [b for b in buildings if b.building_type.find("MFH") != -1]
        number_ap_sfh = sum(1 for b in sfhs for e in b.get_lower_entities() if type(e) == Apartment)
        number_ap_mfh = sum(1 for b in mfhs for e in b.get_lower_entities() if type(e) == Apartment)
        assert len(sfhs) == number_ap_sfh == 5
        assert len(mfhs) == 2
        for d_id, share in self.dd.items():
            if d_id in ["FL", "DL", "EV"]:
                amount_sfh = round(share * number_ap_sfh)
                amount_mfh = round(share * number_ap_mfh)
            elif d_id in ["BAT", "PV"]:
                amount_sfh = round(share * 5)
                amount_mfh = round(share * 2)
            else:
                raise ValueError("Unknown Type {}".format(d_id))
            self.assertEqual(amount_sfh, sum(1 for b in sfhs for e in b.get_entities() if
                                         type(e) == all_entities[d_id]))
            self.assertEqual(amount_mfh, sum(1 for b in mfhs for e in b.get_entities() if
                                         type(e) == all_entities[d_id]))

        op = CentralOptimization(district)
        op.solve(beta=0)
        return


class TestWriteSchedules(unittest.TestCase):
    def setUp(self):
        t = Timer(op_horizon=2)
        p = Prices(t)
        w = Weather(t)
        e = Environment(t, w, p)
        self.cd = CityDistrict(e)
        self.bd = Building(e)
        self.cd.p_el_schedule[:] = [5, 10]
        self.bd.load_schedule("ref")
        self.bd.p_el_schedule[:] = [2, 8]
        return

    def test_to_dict_1(self):
        d = schedule_to_dict([self.cd, self.bd], "default")
        self.assertSetEqual(set([str(self.cd), str(self.bd)]), set(d.keys()))
        d_cd = d[str(self.cd)]
        self.assertSetEqual(set(["default"]), set(d_cd.keys()))
        d_cd_d = d_cd["default"]
        self.assertSetEqual(set(["p_el"]), set(d_cd_d.keys()))
        assert np.allclose([5, 10], d_cd_d["p_el"])
        d_bd = d[str(self.bd)]
        self.assertSetEqual(set(["default"]), set(d_bd.keys()))
        d_bd_d = d_bd["default"]
        self.assertSetEqual(set(["p_el", "p_th_heat", "p_th_cool"]), set(d_bd_d.keys()))
        assert np.allclose([0, 0], d_bd_d["p_el"])
        assert np.allclose([0, 0], d_bd_d["p_th_heat"])
        assert np.allclose([0, 0], d_bd_d["p_th_cool"])
        return

    def test_to_dict_2(self):
        d = schedule_to_dict([self.cd, self.bd])
        self.assertSetEqual(set([str(self.cd), str(self.bd)]), set(d.keys()))
        d_cd = d[str(self.cd)]
        self.assertSetEqual(set(["default"]), set(d_cd.keys()))
        d_cd_d = d_cd["default"]
        self.assertSetEqual(set(["p_el"]), set(d_cd_d.keys()))
        assert np.allclose([5, 10], d_cd_d["p_el"])
        d_bd = d[str(self.bd)]
        self.assertSetEqual(set(["ref"]), set(d_bd.keys()))
        d_bd_d = d_bd["ref"]
        self.assertSetEqual(set(["p_el", "p_th_heat", "p_th_cool"]), set(d_bd_d.keys()))
        assert np.allclose([2, 8], d_bd_d["p_el"])
        assert np.allclose([0, 0], d_bd_d["p_th_heat"])
        assert np.allclose([0, 0], d_bd_d["p_th_cool"])
        return

    def test_to_dict_3(self):
        d = schedule_to_dict([self.cd, self.bd], ["default", "ref"])
        self.assertSetEqual(set([str(self.cd), str(self.bd)]), set(d.keys()))
        d_cd = d[str(self.cd)]
        self.assertSetEqual(set(["default", "ref"]), set(d_cd.keys()))
        d_cd_d = d_cd["default"]
        self.assertSetEqual(set(["p_el"]), set(d_cd_d.keys()))
        assert np.allclose([5, 10], d_cd_d["p_el"])
        d_cd_d = d_cd["ref"]
        self.assertSetEqual(set(["p_el"]), set(d_cd_d.keys()))
        assert np.allclose([0, 0], d_cd_d["p_el"])
        d_bd = d[str(self.bd)]
        self.assertSetEqual(set(["default", "ref"]), set(d_bd.keys()))
        d_bd_d = d_bd["default"]
        self.assertSetEqual(set(["p_el", "p_th_heat", "p_th_cool"]), set(d_bd_d.keys()))
        assert np.allclose([0, 0], d_bd_d["p_el"])
        assert np.allclose([0, 0], d_bd_d["p_th_heat"])
        assert np.allclose([0, 0], d_bd_d["p_th_cool"])
        d_bd_d = d_bd["ref"]
        self.assertSetEqual(set(["p_el", "p_th_heat", "p_th_cool"]), set(d_bd_d.keys()))
        assert np.allclose([2, 8], d_bd_d["p_el"])
        assert np.allclose([0, 0], d_bd_d["p_th_heat"])
        assert np.allclose([0, 0], d_bd_d["p_th_cool"])
        return

    def test_schedule_to_json(self):
        with tempfile.TemporaryDirectory() as tmpdirname:
            schedule_to_json([self.cd, self.bd], op.join(tmpdirname, "test"), ["default", "ref"])
            self.assertTrue(op.exists(op.join(op.join(tmpdirname, "test.json"))))
            schedule_to_json([self.cd], op.join(tmpdirname, "test2.json"), "default")
            self.assertTrue(op.exists(op.join(op.join(tmpdirname, "test2.json"))))
            with open(op.join(tmpdirname, "test3.json"), "w") as test3:
                schedule_to_json([self.cd], test3, "default")
            self.assertTrue(op.exists(op.join(op.join(tmpdirname, "test3.json"))))
            self.assertFalse(filecmp.cmp(op.join(tmpdirname, "test.json"),
                                         op.join(op.join(tmpdirname, "test2.json")),
                                         shallow=False))
            self.assertTrue(filecmp.cmp(op.join(tmpdirname, "test2.json"),
                                        op.join(op.join(tmpdirname, "test3.json")),
                                        shallow=False))
        return

    def test_schedule_to_csv(self):
        with tempfile.TemporaryDirectory() as tmpdirname:
            schedule_to_csv([self.cd, self.bd], op.join(tmpdirname, "test"), ';',
                            ["default", "ref"])
            self.assertTrue(op.exists(op.join(op.join(tmpdirname, "test.csv"))))
            schedule_to_csv([self.cd], op.join(tmpdirname, "test2.csv"), ';', "default")
            self.assertTrue(op.exists(op.join(op.join(tmpdirname, "test2.csv"))))
            schedule_to_csv([self.cd], op.join(tmpdirname, "test3.csv"), ',', "default")
            self.assertTrue(op.exists(op.join(op.join(tmpdirname, "test3.csv"))))
            self.assertFalse(filecmp.cmp(op.join(tmpdirname, "test.csv"),
                                         op.join(op.join(tmpdirname, "test2.csv")),
                                         shallow=False))
            self.assertTrue(filecmp.cmp(op.join(tmpdirname, "test2.csv"),
                                        op.join(op.join(tmpdirname, "test3.csv")),
                                        shallow=False))
        return


class TestPlotSchedules(unittest.TestCase):
    def setUp(self):
        t = Timer(op_horizon=2)
        p = Prices(t)
        w = Weather(t)
        e = Environment(t, w, p)
        self.cd = CityDistrict(e)
        self.bd = Building(e)
        self.cd.addEntity(entity=self.bd, position=[0, 0])
        self.cd.p_el_schedule[:] = [5, 10]
        self.bd.load_schedule("ref")
        self.bd.p_el_schedule[:] = [2, 8]
        self.ap = Apartment(e)
        self.bd.addEntity(self.ap)
        return

    def test_plot_into_dir(self):
        with tempfile.TemporaryDirectory() as tmpdirname:
            plot_entity_directory(self.cd, schedule=["default", "ref"], directory_path=op.join(tmpdirname, "test"),
                                  levels=0)
            cd_set = set([str(self.cd) + ".png"])
            bd_set = cd_set | set([str(self.bd)])
            ap_set = set([str(self.bd) + ".png", str(self.ap) + ".png"])
            self.assertSetEqual(cd_set, set(os.listdir(op.join(tmpdirname, "test"))))

            plot_entity_directory(self.cd, schedule="default", directory_path=op.join(tmpdirname, "test2"),
                                  levels=1)
            self.assertSetEqual(cd_set | set([str(self.bd) + ".png"]), set(os.listdir(op.join(tmpdirname, "test2"))))

            plot_entity_directory(self.cd, schedule="default", directory_path=op.join(tmpdirname, "test3"),
                                  levels=2)
            self.assertSetEqual(bd_set, set(os.listdir(op.join(tmpdirname, "test3"))))
            self.assertSetEqual(ap_set, set(os.listdir(op.join(tmpdirname, "test3", str(self.bd)))))

            plot_entity_directory(self.cd, schedule="default", directory_path=op.join(tmpdirname, "test4"))
            self.assertSetEqual(bd_set, set(os.listdir(op.join(tmpdirname, "test4"))))
            self.assertSetEqual(ap_set, set(os.listdir(op.join(tmpdirname, "test4", str(self.bd)))))

            files = list(cd_set)
            files.extend((op.join(str(self.bd), p) for p in ap_set))
            match, missmatch, errors = filecmp.cmpfiles(op.join(tmpdirname, "test3"), op.join(tmpdirname, "test4"),
                                                        files, shallow=False)

            self.assertListEqual(files, match)
            self.assertListEqual([], missmatch)
            self.assertListEqual([], errors)
        return


class TestSubpackage(unittest.TestCase):

    def test_compute_profile(self):
        t = TimerStub()

        with self.assertRaises(ValueError):
            util.compute_profile(t, [1, 0])
        with self.assertRaises(ValueError):
            util.compute_profile(t, [], 'hmm')
        profile = util.compute_profile(t, [0, 1, 0, 1])
        self.assertTrue(np.array_equal([0, 1, 0, 1], profile))
        daily = [0]*11 + [1, 0]*2 + [1]*9
        profile = util.compute_profile(t, daily, 'daily')
        self.assertTrue(np.array_equal([0, 1, 0, 1], profile))
        weekly = [0]*24 + [0]*11 + [1, 0]*2 + [1]*9 + [0]*120
        profile = util.compute_profile(t, weekly, 'weekly')
        self.assertTrue(np.array_equal([0, 1, 0, 1], profile))
        return

    def test_value_extraction(self):
        v = pyomo.Var(domain=pyomo.Reals)
        with self.assertRaises(ValueError):
            util.extract_pyomo_values(v)

        m = pyomo.ConcreteModel()

        m.v = pyomo.Var(domain=pyomo.Reals)
        m.c = pyomo.Constraint(expr=m.v == 1.0)
        solve_model(m)
        self.assertFalse(m.v.stale)
        e = util.extract_pyomo_values(m.v)
        self.assertEqual(1, e)
        self.assertIs(float, type(e))

        m = pyomo.ConcreteModel()
        m.a = pyomo.Var(domain=pyomo.Reals)
        m.c = pyomo.Constraint(expr=m.a == 1.0)
        m.v = pyomo.Var(pyomo.RangeSet(1, 2), domain=pyomo.Reals)
        solve_model(m)

        m.v[1].value = 1
        m.v[2].value = 2
        # setting the value now removes the stale flag
        self.assertFalse(m.v[1].stale)
        self.assertFalse(m.v[2].stale)
        e = util.extract_pyomo_values(m.v)
        self.assertEqual(1, e[0])
        self.assertEqual(2, e[1])
        self.assertEqual('f', e.dtype.kind)

        m = pyomo.ConcreteModel()
        m.a = pyomo.Var(domain=pyomo.Reals)
        m.c = pyomo.Constraint(expr=m.a == 1.0)
        m.v = pyomo.Var(domain=pyomo.Integers, bounds=(4, 5))
        solve_model(m)
        with pytest.warns(UserWarning):
            e = util.extract_pyomo_values(m.v, verbose=True)
        self.assertEqual(4, e)

        m = pyomo.ConcreteModel()
        m.a = pyomo.Var(domain=pyomo.Reals)
        m.c = pyomo.Constraint(expr=m.a == 1.0)
        m.v = pyomo.Var(domain=pyomo.Integers, bounds=(4, 5))
        m.c2 = pyomo.Constraint(expr=m.v == 4.0)
        solve_model(m)
        m.del_component("c2")
        solve_model(m)
        with pytest.warns(UserWarning):
            e = util.extract_pyomo_values(m.v, verbose=True)
        self.assertEqual(4, e)

        m = pyomo.ConcreteModel()
        m.a = pyomo.Var(domain=pyomo.Reals)
        m.c = pyomo.Constraint(expr=m.a == 1.0)
        m.v = pyomo.Var(domain=pyomo.Integers, bounds=(4.1, 4.3))
        solve_model(m)
        with self.assertRaises(SchedulingError):
            with pytest.warns(UserWarning):
                util.extract_pyomo_values(m.v, verbose=True)

        m = pyomo.ConcreteModel()
        m.a = pyomo.Var(domain=pyomo.Reals)
        m.c = pyomo.Constraint(expr=m.a == 1.0)
        m.v = pyomo.Var(pyomo.RangeSet(1, 2), domain=pyomo.Binary, bounds=(0.1, 0.9))
        solve_model(m)
        with self.assertRaises(SchedulingError):
            with pytest.warns(UserWarning):
                util.extract_pyomo_values(m.v, verbose=True)

        m = pyomo.ConcreteModel()
        m.a = pyomo.Var(domain=pyomo.Reals)
        m.c = pyomo.Constraint(expr=m.a == 1.0)
        m.v = pyomo.Var(domain=pyomo.Integers, bounds=(3.9, 4.3))
        solve_model(m)
        with pytest.warns(UserWarning):
            e = util.extract_pyomo_values(m.v, verbose=True)
        self.assertEqual(4, e)
        self.assertIs(int, type(e))

        m = pyomo.ConcreteModel()
        m.a = pyomo.Var(domain=pyomo.Reals)
        m.c = pyomo.Constraint(expr=m.a == 1.0)
        m.v = pyomo.Var(domain=pyomo.Integers)
        solve_model(m)
        with pytest.warns(UserWarning):
            e = util.extract_pyomo_values(m.v, verbose=True)
        self.assertEqual(0, e)
        self.assertIs(int, type(e))

        m = pyomo.ConcreteModel()
        m.a = pyomo.Var(domain=pyomo.Reals)
        m.c = pyomo.Constraint(expr=m.a == 1.0)
        m.v = pyomo.Var(domain=pyomo.Integers, bounds=(None, -5))
        solve_model(m)
        with pytest.warns(UserWarning):
            e = util.extract_pyomo_values(m.v, verbose=True)
        self.assertEqual(-5, e)
        self.assertIs(int, type(e))

        m = pyomo.ConcreteModel()
        m.a = pyomo.Var(domain=pyomo.Reals)
        m.c = pyomo.Constraint(expr=m.a == 1.0)
        m.v = pyomo.Var(domain=pyomo.Integers, bounds=(-10, -2))
        solve_model(m)
        with pytest.warns(UserWarning):
            e = util.extract_pyomo_values(m.v, verbose=True)
        self.assertEqual(-2, e)
        self.assertIs(int, type(e))

        m = pyomo.ConcreteModel()
        m.a = pyomo.Var(domain=pyomo.Reals)
        m.c = pyomo.Constraint(expr=m.a == 1.0)
        m.v = pyomo.Var(domain=pyomo.Integers, bounds=(2, 10))
        solve_model(m)
        with pytest.warns(UserWarning):
            e = util.extract_pyomo_values(m.v, verbose=True)
        self.assertEqual(2, e)
        self.assertIs(int, type(e))

        m = pyomo.ConcreteModel()
        m.a = pyomo.Var(domain=pyomo.Reals)
        m.c = pyomo.Constraint(expr=m.a == 1.0)
        m.v = pyomo.Var(pyomo.RangeSet(1, 2), domain=pyomo.Binary)
        solve_model(m)
        with pytest.warns(UserWarning):
            e = util.extract_pyomo_values(m.v, verbose=True)
        self.assertEqual(0, e[0])
        self.assertEqual(0, e[1])
        self.assertEqual('b', e.dtype.kind)

        m = pyomo.ConcreteModel()
        m.a = pyomo.Var(domain=pyomo.Reals)
        m.c = pyomo.Constraint(expr=m.a == 1.0)
        m.v = pyomo.Var(pyomo.RangeSet(1, 2), domain=pyomo.Binary, bounds=(1, None))
        solve_model(m)
        with pytest.warns(UserWarning):
            e = util.extract_pyomo_values(m.v, verbose=True)
        self.assertEqual(1, e[0])
        self.assertEqual(1, e[1])
        self.assertEqual('b', e.dtype.kind)
        return


class TimerStub:
    def __init__(self):
        self.time_discretization = 3600
        self.simu_horizon = 4

    @staticmethod
    def time_in_week(from_init):
        if not from_init:
            raise ValueError
        return 36

    @staticmethod
    def time_in_day(from_init):
        if not from_init:
            raise ValueError
        return 12


def solve_model(model):
    # hack to suppress pyomo no constraint warning
    if not hasattr(model, "simple_var"):
        model.simple_var = pyomo.Var(domain=pyomo.Reals, bounds=(None, None), initialize=0)
        model.simple_constr = pyomo.Constraint(expr=model.simple_var == 1)
    opt = pyomo.SolverFactory(solvers.DEFAULT_SOLVER, **solvers.DEFAULT_SOLVER_OPTIONS.get("__call__", {}))
    if isinstance(opt, PersistentSolver):
        opt.set_instance(model, **solvers.DEFAULT_SOLVER_OPTIONS.get("set_instance", {}))
    return opt.solve(model, **solvers.DEFAULT_SOLVER_OPTIONS.get("solve", {}))